#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  3 18:13:45 2024

@author: ncro8394
"""

from numpy import (append, array, asarray, concatenate, cumsum, diff, inf, 
                   insert, isnan, median, multiply, nan, nanmean, nanstd, pad, 
                   percentile, repeat, sqrt, squeeze, where)
from os import listdir, mkdir, path, getpid
from wonambi import Dataset, graphoelement
from wonambi.attr import Annotations, create_empty_annotations
from wonambi.detect.spindle import transform_signal
from wonambi.trans import fetch
import mne
from scipy.signal import butter, correlate, filtfilt, find_peaks, peak_widths
import yasa
from copy import deepcopy
from ..utils.logs import create_logger
from ..utils.load import load_channels, load_sessions, rename_channels
from ..utils.misc import (merge_events, reconstruct_stitches, remove_event, 
                          remove_duplicate_evts)
import gc
import psutil

def print_memory_usage(note="", logger = create_logger("Memory check")):
    process = psutil.Process(getpid())
    mem_mb = process.memory_info().rss / 1024 ** 2  # in MB
    logger.debug(f"[{note}] Memory usage: {mem_mb:.2f} MB")

class SAND:
    
    """ Seapipe Artefact and Noise Detection (S.A.N.D)

        This module runs automated artefact detection with the option of using
        previously published staging algorithms:
            1. YASA (standard deviation)
            2. YASA (covariance)
            3. Seapipe artefact detection
        
    """   
    
    def __init__(self, rec_dir, xml_dir, out_dir, eeg_chan, ref_chan,
                 rater = None, grp_name = 'eeg', 
                 subs='all', sessions='all', tracking = None):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.eeg_chan = eeg_chan
        self.ref_chan = ref_chan
        self.rater = rater
        self.grp_name = grp_name
        
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking


    def detect_artefacts(self, method, label = "individual", win_size = 5,
                               filetype = '.edf', 
                               stage = ['NREM1', 'NREM2', 'NREM3', 'REM'],
                               logger = create_logger('Detect artefacts')):
        
        ''' Automatically detects artefacts.
        
            Creates a new annotations file if one doesn't already exist.
        
        INPUTS:
            
            method      ->   str of name of automated detection algorithm to 
                             detect staging with. 
                             Current methods supported: 
                                 1. 'Vallat2021' (https://doi.org/10.7554/eLife.70092)
                                 2. 'seapipe' 
                             
            qual_thresh ->   Quality threshold. Any stages with a confidence of 
                             prediction lower than this threshold will be set 
                             to 'Undefined' for futher manual review.
   
        
        '''
        print_memory_usage("Start")
        
        ### 0.a Set up logging
        flag = 0
        tracking = self.tracking
        if label == "allchans":
            chan_msg = "all channels at once."
        else:
            chan_msg = "each channel individually."
            
        logger.info('')
        logger.debug(rf"""Commencing artefact detection... 
                     
                                             ____
                                      /^\   / -- )
                                     / | \ (____/
                                    / | | \ / /
                                   /_|_|_|_/ /
                                    |     / /
                     __    __    __ |    / /__    __    __
                    [  ]__[  ]__[  ].   / /[  ]__[  ]__[  ]     ......
                    |__            ____/ /___           __|    .......
                       |          / .------  )         |     ..........
                       |         / /        /          |    ............
                       |        / /        / _         |  ...............
                   ~._..-~._,….-ˆ‘ˆ˝\_,~._;––' \_.~.~._.~'\................  
                       
            
                    Seapipe Artefact and Noise Detection
                    (S.A.N.D)

                    Method: {method}
                    
                    Applying to: {chan_msg} 
                    
                                                    """,)
        ### 1. First we check the directories
        # a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug("Output directory: " + self.out_dir + " exists")
        else:
            mkdir(self.out_dir)
        
        # b. Check input list
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [p for p in subs if not '.' in p]
        else:
            logger.error("'subs' must either be an array of subject ids or = 'all' ")  
            return
        
        ### 2. Begin loop through dataset
       
        # a. Begin loop through participants
        subs.sort()
        for i, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2)
            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'slowosc':{}} 
                
                # Define recording
                rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)][0]
                except:
                    logger.warning(f'No input {filetype} file in {rdir}')
                    flag += 1
                    break
                
                ## f. Channel setup 
                pflag = deepcopy(flag)
                flag, chanset = load_channels(sub, ses, self.eeg_chan, 
                                              self.ref_chan, flag, logger)
                if flag - pflag > 0:
                    logger.warning(f'Skipping {sub}, {ses}...')
                    break
                
                newchans = rename_channels(sub, ses, self.eeg_chan, logger) 
                
                # Check if applying to all channels or chan-by-chan
                if label == "allchans":
                    def _freeze_ref(val):
                        """Recursively convert nested iterables to tuples for hashing."""
                        if isinstance(val, (list, tuple)):
                            return tuple(_freeze_ref(v) for v in val)
                        if hasattr(val, "tolist"):
                            return _freeze_ref(val.tolist())
                        return val
                    # Check if references are the same for each channel
                    ref_chans = {_freeze_ref(val) for val in chanset.values()}
                    # If not, setup for running per channel
                    if len(ref_chans) > 1:
                        logger.warning("Channel setup 'all_chans' was set 'True', but "
                                       f"Channel:Reference pairings are unique for {sub}, "
                                       f"{ses}. Therefore, we'll detect artefacts PER CHANNEL.")
                        flag += 1
                        label = "individual"
                
                # Get data 
                dset = Dataset(rdir + edf_file)
                s_freq = int(dset.header['s_freq'])
                
                # d. Load/create for annotations file
                if not path.exists(self.xml_dir + '/' + sub):
                    mkdir(self.xml_dir + '/' + sub)
                if not path.exists(self.xml_dir + '/' + sub + '/' + ses):
                     mkdir(self.xml_dir + '/' + sub + '/' + ses)
                xdir = self.xml_dir + '/' + sub + '/' + ses
                
                xml_file = [x for x in listdir(f'{xdir}') if '.xml' in x]
                if len(xml_file) > 1:
                    logger.warning(f'More than 1 annotations file found for '
                                   f'{sub}, {ses} in {xdir}. Skipping...')
                    continue
                if len(xml_file) < 1:
                    logger.warning(f'No annotations file was found for '
                                   f'{sub}, {ses} in {xdir}. Skipping...')
                    continue
                else:
                    xml_file = f'{xdir}/{xml_file[0]}'
                    
                if not path.exists(xml_file):
                    create_empty_annotations(xml_file, dset)
                    logger.warning(f"No annotations file exists. Creating " 
                                   f"annotations file for {sub}, {ses} and" 
                                   "detecting Artefacts WITHOUT hypnogram.")
                    annot = Annotations(xml_file)
                    hypno = None
                else:
                    logger.debug(f'Annotations file exists for {sub}, {ses},'
                                 'staging will be used for Artefact detection.')
                    logger.warning('Any previously detected artefacts will be '
                                   'overwritten.')

                    # Extract hypnogram
                    annot = Annotations(xml_file)
                    hypno = [x['stage'] for x in annot.get_epochs()]
                    stage_key = {'Wake':0,
                                 'NREM1':1,
                                 'NREM2':2,
                                 'NREM3':3,
                                 'REM':4,
                                 'Undefined':0,
                                 'Unknown':0,
                                 'Artefact':0}
                    
                    hypno = array([int(stage_key[x]) for x in hypno])
                
                evts = []
                for chan in chanset:
                    if newchans:
                        logger.debug(f'Detecting for {newchans[chan]} : {chanset[chan]}')
                    else:
                        logger.debug(f'Detecting for {chan} : {chanset[chan]}')
                    
                    remove_event(annot, 'Artefact', chan)
                    
                    if 'yasa' in method:     
                        
                        ## c. Load recording
                        try:
                            raw = mne.io.read_raw_edf(rdir + edf_file, 
                                                      include = chan + chanset[chan],
                                                      preload=True, verbose = False)
                            
                            mne.set_eeg_reference(raw, ref_channels = chanset[chan])
                                                  
                            s_freq = raw.info["sfreq"]
                        except Exception as e:
                            logger.warning(f'Error loading {filetype} file in {rdir}, {repr(e)}')
                            flag += 1
                            break
                        
                        yasa_meth = 'covar' if 'covar' in method else 'std'
                            
                        # Convert raw data to array    
                        data = raw.to_data_frame()
                        inds = [x for x in data if x in chan]
                        data = data[inds].T
                        data = data.to_numpy()
                        
                        # Upsample hypnogram to match raw data
                        hypno_up = yasa.hypno_upsample_to_data(hypno, 1/30, data, 
                                                               sf_data=s_freq)
                        
                        # Detect artefacts
                        n_chan_reject = 1 if data.shape[0] == 1 else 2
                        art, zscores = yasa.art_detect(data, s_freq, 
                                                       window = win_size, 
                                                       hypno = hypno_up, 
                                                       include = (1, 2, 3, 4), 
                                                       method = yasa_meth, 
                                                       threshold = 3, 
                                                       n_chan_reject = n_chan_reject, 
                                                       verbose = False)
                        
                        # Upsample artefacts to match raw data
                        art = multiply(art, 1)
                        sf_art = 1/win_size
                        art_up = yasa.hypno_upsample_to_data(art, sf_art, data, 
                                                             s_freq)
                        
                        # Find start/end times of artefacts
                        peaks = find_peaks(art_up)
                        properties = peak_widths(art_up, peaks[0])
                        times = [x for x in zip(properties[2],properties[3])]
        
                    elif 'seapipe' in method:
                        
                        logger.debug('Loading Data..')

                        segments = fetch(dset, annot, cat = (1,1,1,1), 
                                         stage=stage)
                        
                        if len(segments) < 1:
                            logger.warning('No valid data found, skipping {chan}')
                            continue
                        
                        # Save stitches (to recompute times later)
                        try:
                            stitches = segments[0]['times']
                        except:
                            logger.warning('No valid data found, skipping {chan}')
                            
                        # Read data
                        segments.read_data(chan, ref_chan = chanset[chan], 
                                           grp_name=self.grp_name)
                        data = segments[0]['data'].data[0]
                        
                        # Filter data above 40Hz
                        dat = transform_signal(data[0], s_freq, 'high_butter', 
                                         method_opt={'freq':40,
                                                     'order':3})
                        
                        
                        ## ---- Step 1. Detect flatlines in data ----
                        logger.debug("Detecting flatlines...")
                        # Align the filtered data back to the raw EEG recording
                        dat_full = reconstruct_stitches(dat, stitches, s_freq,
                                                        replacement=nan)
                        
                        # Let's setup a function to detect flatlines
                        def detect_flatlines(ts, s_freq, tolerance=1e-5, min_length=5):
                            diffs = abs(diff(ts))
                            is_flat = diffs < tolerance
                        
                            # Find indices where flatlines start and stop
                            change_points = where(diff(concatenate(([False], 
                                                                     is_flat, 
                                                                     [False]))))[0]
                            starts, ends = change_points[::2], change_points[1::2]
                        
                            # Filter out short flatlines and adjust for s_freq offset
                            flatlines = [(max(0, start - int(s_freq / 4)), min(len(ts), 
                                                                               end + int(s_freq / 4))) 
                                         for start, end in zip(starts, ends) if 
                                         (end - start) >= min_length]
                        
                            return flatlines 
                        flatlines = detect_flatlines(dat_full, s_freq)
                        
                        # Force garbage collection to free memory
                        del dat_full
                        gc.collect()  
                        
                        
                        ## ---- Step 2. Find large shifts in frequencies (>40Hz) ----
                        logger.debug("Detecting movement artefacts...")
                        # Mask out negative values
                        dat[dat<0] = 0
                        
                        # Calculate sliding RMS
                        dat = transform_signal(dat, s_freq, 'moving_rms',
                                         method_opt = {'dur':win_size,
                                                       'step':1})
                        
                        # Put back all the samples that were averaged within the window 
                        dat = repeat(dat, s_freq)

                        # Filter (smooth) the RMS
                        dat = transform_signal(dat, s_freq, 'low_butter', 
                                         method_opt={'freq':0.5,
                                                     'order':3})

                        # Align this filtered data back to the raw EEG recording 
                        dat_full = reconstruct_stitches(dat, stitches, s_freq,
                                                        replacement = nan)
                        
                        # Threshold and detect artefact 'events'
                        #threshold = percentile(dat, 97) # another option (future)
                        threshold = nanstd(dat_full)*2.5
                        dat_full[dat_full<threshold] = 0
                        dat_full[dat_full>threshold] = 1
                        movements = detect_above_zero_regions(dat_full)
                        
                        # Force garbage collection to free memory
                        del dat, dat_full
                        gc.collect()
                        
                        
                        # ---- Step 3. Large deflections in signal ----
                        logger.debug("Detecting deflection artefacts...")
                        
                        dat = append(0, squeeze(diff(data)))
                        dat_full = reconstruct_stitches(dat, stitches, s_freq,
                                                         replacement=nan)
                        threshold = 10 * nanstd(dat_full)
                        dat_full[dat_full<threshold] = 0
                        dat_full[dat_full>=threshold] = 1
                        deflections = detect_above_zero_regions(dat_full)
                        deflections = [(x[0] - int(s_freq/2), 
                                        x[1] + int(s_freq/2)) for x in deflections]

                        
                        ## TODO: Step 3. Let's look for sharp slow waves in REM
                        # if 'REM' in stage:
                        #     logger.debug("Detecting eye movement artefacts in REM...")
                        #     segments = fetch(dset, annot, cat = (1,1,1,1), 
                        #                      stage=['REM'])
                            
                        #     # Save stitches (to recompute times later)
                        #     stitches = segments[0]['times']
                        #     # Read data
                        #     segments.read_data(chan, ref_chan=ref, 
                        #                        grp_name=self.grp_name)
                        #     data = segments[0]['data'].data[0]
                            
                        #     # Filter (smooth) the RMS
                        #     dat = transform_signal(data[3], s_freq, 'low_butter', 
                        #                      method_opt={'freq':5,
                        #                                  'order':3})
                            
                        #     dat = transform_signal(dat, s_freq, 'moving_sd',
                        #                      method_opt = {'dur':3,
                        #                                    'step':1})
                            
                        #     # Put back all the samples that were averaged within the window 
                        #     dat = repeat(dat, s_freq)
                            
                            
                        #     # Align this filtered data back to the raw EEG recording
                        #     dat_reconstructed = reconstruct_stitches(dat, 
                        #                                               stitches, 
                        #                                               s_freq)
                            
                            
                        #     # Threshold and detect artefact 'events'
                        #     threshold = 50
                        #     dat_reconstructed[dat_reconstructed<50] = 0
                        #     dat_reconstructed[dat_reconstructed>50] = 1
                        #     REMS = detect_above_zero_regions(dat_reconstructed)
                            

                        ## ---- Step 4. Combine all artefact types ----
                        merged = flatlines + deflections +  movements
                        merged = merge_segments(merged)
                        
                        # Convert times back into seconds for annotations
                        times = [(x / s_freq, y / s_freq) for x, y in merged]

                        # Force garbage collection to free memory
                        del data, dat, dat_full, segments
                        gc.collect()

                    else:
                        logger.critical("Currently the only methods that are" 
                                        " functioning include:"
                                        "\n'yasa_std', 'yasa_covar', or 'seapipe'."
                                        ) 
                        return
                    
                    # ---- Append all events in master list ----
                    if label == "allchans":
                        channame = ['']
                    else:
                        channame = f'{chan} ({self.grp_name})'
                        
                    for x in times:
                        if x[1] - x[0] > 400:
                            logger.warning('Artefact >400s detected. Review.')
                        
                        evts.append({'name':'Artefact',
                                     'start':float(x[0]),
                                     'end':float(x[1]),
                                     'chan':channame,
                                     'stage':'',
                                     'quality':'Good',
                                     'cycle':''})
                    
                    del times
                    gc.collect()
                    
                # ---- Save events to Annotations file ----
                grapho = graphoelement.Graphoelement()
                grapho.events = evts          
                grapho.to_annot(annot)
                
                del evts, dset
                gc.collect()
                
                # Remove duplicates
                if label == 'allchans':
                    merge_events(annot, 'Artefact')
                    remove_duplicate_evts(annot, 'Artefact')
                elif label == 'individual':
                    for chan in chanset:
                        merge_events(annot, 'Artefact', 
                                     chan = f'{chan} ({self.grp_name})')
                        remove_duplicate_evts(annot, 'Artefact', 
                                              chan = f'{chan} ({self.grp_name})')
                        
                
        return
    
    
    
def detect_ecg_artifacts(eeg_signal, sfreq, bandpass=(1, 40), template_window=0.6,
                         peak_percentile=95, match_percentile=99, min_rr_interval=0.5):
    """
    Automatically detects ECG artifacts in a single EEG channel using template matching.

    Parameters:
    - eeg_signal: 1D numpy array of EEG data
    - sfreq: Sampling frequency in Hz
    - bandpass: Tuple of (low, high) bandpass filter cutoff in Hz
    - template_window: Length of ECG template window in seconds
    - peak_percentile: Percentile threshold for peak detection
    - match_percentile: Correlation percentile threshold for match detection
    - min_rr_interval: Minimum interval between matches in seconds

    Returns:
    - clean_matches: Sample indices of detected ECG artifacts
    - template: The derived ECG template used for matching
    - corr: Cross-correlation trace
    """
    
    
    # 1. Bandpass filter
    def bandpass_filter(data, low, high, fs, order=4):
        nyq = 0.5 * fs
        b, a = butter(order, [low / nyq, high / nyq], btype='band')
        return filtfilt(b, a, data)

    filtered = bandpass_filter(eeg_signal, *bandpass, fs=sfreq)

    # 2. Detect peaks likely to be ECG R-peaks
    min_distance = int(min_rr_interval * sfreq)
    height_thresh = percentile(filtered, peak_percentile)
    peaks, _ = find_peaks(filtered, 
                          height = height_thresh, 
                          distance = min_distance)

    # 3. Extract candidate segments around each peak
    win_len = int(template_window * sfreq)
    half_win = win_len // 2
    segments = []

    for peak in peaks:
        if peak - half_win >= 0 and peak + half_win < len(filtered):
            segment = filtered[peak - half_win:peak + half_win]
            segments.append(segment)

    if not segments:
        raise ValueError("No valid ECG segments found. Try lowering peak_percentile or check EEG signal.")

    segments = array(segments)

    # 4. Create the ECG template
    template = median(segments, axis=0)

    # 5. Cross-correlate the template with the EEG signal
    corr = fast_normalized_cross_correlation(filtered, template)
    threshold = percentile(corr[~isnan(corr)], match_percentile)
    matches = where(corr > threshold)[0]

    # 6. Prune close matches
    pruned = []
    last = -inf
    for m in matches:
        if m - last > min_distance:
            pruned.append(m)
            last = m

    clean_matches = array(pruned)

    return clean_matches, template, corr    
    

def fast_normalized_cross_correlation(signal, template):
    """
    Efficient normalized cross-correlation using cumulative sums.

    Parameters:
    - signal: 1D array
    - template: 1D array (shorter than signal)

    Returns:
    - norm_corr: Normalized cross-correlation (same length as signal)
    """
    signal = asarray(signal)
    template = asarray(template)
    template_len = len(template)
    template = (template - nanmean(template)) / nanstd(template)

    # Raw cross-correlation (same length as signal)
    raw_corr = correlate(signal, template, mode='same')

    # Compute running mean and std of signal (same window length as template)
    window = template_len
    runmean = cumsum(insert(signal, 0, 0))
    runmean2 = cumsum(insert(signal**2, 0, 0))

    sum_window = runmean[window:] - runmean[:-window]
    sum2_window = runmean2[window:] - runmean2[:-window]

    mean_window = sum_window / window
    std_window = sqrt(sum2_window / (window - mean_window**2))

    # Pad to match length of signal
    std_padded = pad(std_window, (window // 2, len(signal) - len(std_window) - 
                                  window // 2), mode='edge')

    # Avoid division by zero
    std_padded[std_padded == 0] = 1e-10

    norm_corr = raw_corr / (std_padded * window)

    return norm_corr



def detect_above_zero_regions(signal):
    """
    Detects regions where the signal is above zero, excluding values
    that are adjacent to NaN periods.

    Parameters:
    - signal: 1D array-like

    Returns:
    - regions: list of tuples (start_idx, end_idx)
    """
    signal = asarray(signal)
    nans = isnan(signal)
    is_above_zero = signal > 0

    # Flag values adjacent to NaNs
    padded_nan = pad(nans, (1, 1), constant_values=False)
    adjacent_to_nan = padded_nan[:-2] | padded_nan[2:]

    # Remove values that are adjacent to NaNs
    clean_mask = is_above_zero & ~adjacent_to_nan

    # Extract contiguous regions
    mask_diff = diff(clean_mask.astype(int))
    starts = where(mask_diff == 1)[0] + 1
    ends = where(mask_diff == -1)[0] + 1

    if clean_mask[0]:
        starts = insert(starts, 0, 0)
    if clean_mask[-1]:
        ends = append(ends, len(clean_mask))

    return list(zip(starts, ends))


# def detect_above_zero_regions(signal):
#     """
#     Detects regions where the signal is above zero.

#     Parameters:
#     - signal: 1D array
#         Input signal where the regions of interest (above zero) are detected.

#     Returns:
#     - regions: list of tuples
#         Each tuple contains the start and end indices of regions where the signal is above zero.
#     """
#     signal = asarray(signal)
#     valid = ~isnan(signal)

#     # Compute diffs and mask invalid transitions
#     diffs = diff(signal.astype(float))
#     valid_diffs = valid[:-1] & valid[1:]

#     raw_starts = where((diffs > 0) & valid_diffs)[0] + 1
#     raw_ends = where((diffs < 0) & valid_diffs)[0] + 1

#     # Edge case: signal starts above 0 and is valid
#     if valid[0] and signal[0] > 0:
#         raw_starts = insert(raw_starts, 0, 0)

#     # Edge case: signal ends above 0 and is valid
#     if valid[-1] and signal[-1] > 0:
#         raw_ends = append(raw_ends, len(signal))

#     # Now, match starts to ends
#     starts = []
#     ends = []
#     s_idx = 0
#     e_idx = 0

#     while s_idx < len(raw_starts) and e_idx < len(raw_ends):
#         if raw_starts[s_idx] < raw_ends[e_idx]:
#             starts.append(raw_starts[s_idx])
#             # find first end after this start
#             while e_idx < len(raw_ends) and raw_ends[e_idx] < raw_starts[s_idx]:
#                 e_idx += 1
#             if e_idx < len(raw_ends):
#                 ends.append(raw_ends[e_idx])
#                 e_idx += 1
#             s_idx += 1
#         else:
#             # skip any unmatched end that comes before the first start
#             e_idx += 1

#     return list(zip(starts, ends))

    

def merge_segments(segments):
    """
    Merges overlapping or adjacent segments in a list of (start, end) tuples.

    Parameters:
      segments (list of tuples): List of (start, end) index pairs.

    Returns:
      list of tuples: Merged (start, end) segments.
    """
    if not segments:
        return []

    # Sort segments by start time
    segments.sort()

    # Initialize merged list with the first segment
    merged = [segments[0]]

    for start, end in segments[1:]:
        prev_start, prev_end = merged[-1]

        if start <= prev_end:  # Overlapping or adjacent segments
            merged[-1] = (prev_start, max(prev_end, end))  # Merge them
        else:
            merged.append((start, end))  # No overlap, add as new segment

    return merged



# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 15:51:24 2021

@author: Nathan Cross
"""
from datetime import datetime, date

from os import listdir, mkdir, path, walk
from . cfc_func import _allnight_ampbin, circ_wwtest, mean_amp, klentropy
from . octopus import pac_method
from seapipe.utils.misc import bandpass_mne, laplacian_mne, notch_mne, notch_mne2
from copy import deepcopy
import shutil
from math import degrees, radians
import mne
import matplotlib.pyplot as plt
from numpy import (angle, append, argmax, array, arange, asarray, ceil, concatenate, 
                   empty, histogram, interp, isnan, linspace, log, logical_and, mean, 
                   median, nan, nanmean, nanmedian, ndarray, newaxis, ones, pi, random, repeat, 
                   reshape, roll, save, sin, size, squeeze, sqrt, std, sum, tile, where, zeros) 
from numpy.matlib import repmat
from pandas import DataFrame, concat, read_csv
from pathlib import Path
from safepickle import dump, load
from pingouin import (circ_mean, circ_r, circ_rayleigh, circ_corrcc, circ_corrcl)
from scipy.signal import hilbert
from scipy.stats import zscore
import sys
from tensorpac import Pac, EventRelatedPac
from wonambi import Dataset
from wonambi.trans import fetch
from wonambi.attr import Annotations 
from wonambi.detect.spindle import transform_signal
from seapipe.utils.logs import create_logger, create_logger_outfile
from ..utils.load import (load_channels, load_adap_bands, rename_channels, 
                          load_sessions, read_inversion, read_manual_peaks)
from ..utils.misc import remove_duplicate_evts
from ..utils.misc import infer_polarity


class pacats:

    def __init__(self, rootpath, rec_dir, xml_dir, out_dir, chan, ref_chan, 
                 grp_name, stage, rater = None, subs = 'all', 
                 sessions = 'all', reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                 tracking = None):
        
        self.rootpath = rootpath
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.stage = stage
        self.rater = rater
        self.reject = reject_artf
        
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking


    def pac_it(self, cycle_idx, cat, nbins, filter_opts, 
                 epoch_opts, frequency_opts, filetype = '.edf',
                 idpac = (2,3,4), min_dur = 1, 
                 adap_bands_phase = 'Fixed', frequency_phase = (0.5,1.25), 
                 adap_bands_amplitude = 'Fixed', frequency_amplitude = (11,16), 
                 adap_bw = 4, invert = False, progress=True,
                 logger = create_logger('Phase-amplitude coupling')):

        '''
        P.A.C.A.T.S
        
        Phase Amplitude Coupling Across Time Series
        
        This script runs Phase Amplitude Coupling analyses on continuous sleep 
        EEG data (unlike O.C.T.O.P.U.S. which is for events). 
        The method for calculating PAC is set by the parameter <idpac>. 
        For more information on the available methods, refer to the documentation of 
        tensorpac (https://etiennecmb.github.io/tensorpac/index.html) or the article
        (Combrisson et al. 2020, PLoS Comp Bio: https://doi.org/10.1371/journal.pcbi.1008302)
        
        The script does the following:
            1. Extracts the continuous EEG signal segments defined by <cat> and 
                if the segment is too short ± a buffer on either side of length 
                (in sec) specified by <buffer>.
            2. For these EEG segments, filters the signal within a given frequency range
               specified by <fpha> to obtain the phase, and again within a given frequency 
               range specified by <famp>.  
            3. FOR EACH EACH EVENT: the instantaneous amplitude of the signal filtered 
               within a given frequency range specified by <famp> will be calculated 
               via the Hilbert transform, and the amplitude will be averaged across a 
               set number of phase bins specified by <nbins>. The phase bin with the 
               maxmimum mean amplitude will be stored.
            4. ACROSS ALL EVENTS: the average phase bin with the maximum amplitude will 
               be calculated (circular mean direction).
            5. The filtered events will also be concatenated and stored in blocks of 50,
               so that the PAC strength (method pecficied by the 1st entry in <idpac>)
               can be calculated AND surrogates can be accurately generated to test for
               the significance of PAC in each participant. The generation of surrogates
               is specified by the 2nd entry in <idpac>, and the correction of PAC 
               strength is also calculated, specified by the 3rd entry in <idpac>.
            6. Other metrics are also calculated for each participant and visit, notably:
                - mean vector length (given from mean circular calculation)
                - correlation between amplitudes (averaged over all events) and the phase 
                   giving sine wave
                - Rayleigh test for non-uniformity of circular data (sig. test for 
                                                                     preferred phase)
                
               
        If laplacian = True then a Laplacian spatial filter will be applied to remove high frequency EMG 
        noise. In this scenario you will need to provide a list of channel names to include in the laplacian
        spatial filtering. 
                        ## WARNING: TEST WHAT THE LAPLACIAN FILTER DOES TO THE POWER SPECTRUM BEFORE USING
                                    THIS OPTION
        
        If adap_bands = (True,True) then the (phase,amplitude) signal will be filtered within an adapted 
        frequency range for each individual subject or recording.
        
        The output provided by this script will be an array of size:
            [#cycles x #bins] (if cycle_idx is not None)
            [1 x #nbins]      (if cycle_idx is None)
                                            - corresponding to the mean amplitude of the signal 
                                             (across all cycles or events) per phase bin.
    
        '''
        # Get method descriptions
        pac_list = pac_method(0, 0, 0, list_methods=True)
        methods = pac_list[0]
        surrogates = pac_list[1]
        corrections = pac_list[2]
    
        tracking = self.tracking
        flag = 0
        
        logger.info('')
        logger.debug(rf"""Commencing phase-amplitude coupling pipeline... 
                                
                     
                                                                       /~~\
                        ____                                         /'o  |
                     .';;|;;\            _,-;;;\;-_               ,'  _/'|
                    `\_/;;;/;\         /;;\;;;;\;;;,             |     .'
                       `;/;;;|      ,;\;;;|;;;|;;;|;\          ,';;\  |
                       |;;;/;:     |;;;\;/~~~~\;/;;;|        ,;;;;;;.'
                      |;/;;;|     |;;;,'      `\;;/;|      /;\;;;;/
                      `|;;;/;\___/;~\;|         |;;;;;----\;;;|;;/'
                       `;/;;;|;;;|;;;,'         |;;;;|;;;;;|;;|/'
                        `\;;;|;;;/;;,'           `\;/;;;;;;|/~'
                         `\/;;/;;;/               `~------'
                           `~~~~~  

                
                Phase Amplitude Coupling Across Time Series
                (P.A.C.A.T.S)
                
                Method: {methods[idpac[0]]}
                Correction: {surrogates[idpac[1]]}
                Normalisation: {corrections[idpac[2]]}
                                  
                                                    """,)
        ### 0.b. Set up organisation of export
        if cat[0] + cat[1] == 2:
            model = 'whole_night'
            logger.debug('Analysing PAC for the whole night.')
        elif cat[0] + cat[1] == 0:
            model = 'stage*cycle'
            logger.debug('Analysing PAC per stage and cycle separately.')
        elif cat[0] == 0:
            model = 'per_cycle'
            logger.debug('Analysing PAC per cycle separately.')
        elif cat[1] == 0:
            model = 'per_stage'  
            logger.debug('Analysing PAC per stage separately.')
        if 'cycle' in model and cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        
        # Log filtering options
        if filter_opts['notch']:
            logger.debug(f"Applying notch filtering: {filter_opts['notch_freq']} Hz")
        if filter_opts['notch_harmonics']: 
            logger.debug('Applying notch harmonics filtering.')
        if filter_opts['bandpass']:
            logger.debug(f"Applying bandpass filtering: {filter_opts['highpass']} - {filter_opts['lowpass']} Hz")
        if filter_opts['laplacian']:
            logger.debug('Applying Laplacian filtering.')
        
        ### 1. First we check the directories
        # a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug("Output directory: " + self.out_dir + " exists")
        else:
            mkdir(self.out_dir)
        
        # b. Check input list
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [p for p in subs if not '.' in p]
        else:
            logger.error("'subs' must either be an array of subject ids or = 'all' ")   
        
        ### 2. Begin loop through dataset
        # a. Begin loop through participants
        subs.sort()
        for i, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2) 
            
            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'pac':{}} 
                
                
                ## c. Load recording
                rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}')
                    break
                
                ## d. Load annotations
                xdir = self.xml_dir + '/' + sub + '/' + ses + '/'
                try:
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]
                    # Copy annotations file before beginning
                    if not path.exists(self.out_dir):
                        mkdir(self.out_dir)
                    if not path.exists(self.out_dir + '/' + sub):
                        mkdir(self.out_dir + '/' + sub)
                    if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                        mkdir(self.out_dir + '/' + sub + '/' + ses)
                    outpath = self.out_dir + '/' + sub + '/' + ses
                    backup_file = (f'{outpath}/{sub}_{ses}_spindle.xml')
                    if not path.exists(backup_file):
                        shutil.copy(xdir + xml_file[0], backup_file)
                    else:
                        logger.debug(f'Using annotations file from: {xdir}')
                except:
                    logger.warning(f' No input annotations file in {xdir}')
                    flag +=1
                    break
                
                # 3.a. Read annotations file
                annot = Annotations(backup_file, rater_name=self.rater)
                
                ## b. Get sleep cycles (if any)
                if cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[y - 1] for y in cycle_idx if y <= len(all_cycles)]
                else:
                    cycle = None
                
                ## c. Channel setup 
                pflag = deepcopy(flag)
                flag, chanset = load_channels(sub, ses, self.chan, self.ref_chan,
                                              flag, logger)
                if flag - pflag > 0:
                    logger.warning(f'Skipping {sub}, {ses}...')
                    flag +=1
                    break
                newchans = rename_channels(sub, ses, self.chan, logger)
                
                # 4.a. Loop through channels
                for c, ch in enumerate(chanset):
                    
                    # b. Rename channel for output file (if required)
                    if newchans:
                        fnamechan = newchans[ch]
                        filter_opts['renames'] = {newchans[ch]:ch}
                        filter_opts['laplacian_rename'] = True
                    else:
                        fnamechan = ch
                        filter_opts['laplacian_rename'] = False

                    if ch == '_REF':
                        filter_opts['laplacian_rename'] = False
                        if not filter_opts['oREF']:
                            try:
                                filter_opts['oREF'] = newchans[ch]
                            except:
                                logger.warning("Channel selected is '_REF' but "
                                               "no information has been given "
                                               "about what standard name applies, "
                                               "either via filter_opts['oREF'] "
                                               "or 'chanset_rename' in the tracking "
                                               "sheet (see user guide). Skipping "
                                               "channel...")
                                flag += 1
                                continue
                    else:
                        filter_opts['oREF'] = None
                        
                    # c. Set frequency bands for:
                    # Phase
                    if adap_bands_phase == 'Fixed':
                        f_pha = frequency_phase   
                    elif adap_bands_phase == 'Manual':
                        f_pha = read_manual_peaks(self.roothpath, sub, ses, ch, 
                                                 adap_bw, logger)
                    elif adap_bands_phase == 'Auto':
                        stagename = '-'.join(self.stage)
                        band_limits = f'{frequency_phase[0]}-{frequency_phase[1]}Hz'
                        f_pha = load_adap_bands(self.tracking['fooof'], sub, ses,
                                               fnamechan, stagename, band_limits, 
                                               adap_bw, logger)
                    if not f_pha:
                        logger.warning('Will use fixed frequency bands for PHASE instead.')
                        f_pha = frequency_phase
                    if not chanset[ch]:
                        logchan = ['(no re-refrencing)']
                    else:
                        logchan = chanset[ch]
                    logger.debug(f"Using PHASE frequency band: {round(f_pha[0],2)}-{round(f_pha[1],2)} Hz for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")    
                    
                    # Amplitude
                    if adap_bands_amplitude == 'Fixed':
                        f_amp = frequency_amplitude   
                    elif adap_bands_amplitude == 'Manual':
                        f_amp = read_manual_peaks(self.roothpath, sub, ses, ch, 
                                                 adap_bw, logger)
                    elif adap_bands_amplitude == 'Auto':
                        stagename = '-'.join(self.stage)
                        band_limits = f'{frequency_amplitude[0]}-{frequency_amplitude[1]}Hz'
                        f_amp = load_adap_bands(self.tracking['fooof'], sub, ses,
                                               fnamechan, stagename, band_limits, 
                                               adap_bw, logger)
                    if not f_amp:
                        logger.warning('Will use fixed frequency bands for PHASE instead.')
                        f_amp = frequency_amplitude
                    if not chanset[ch]:
                        logchan = ['(no re-refrencing)']
                    else:
                        logchan = chanset[ch]
                    logger.debug(f"Using AMPLITUDE frequency band: {round(f_amp[0],2)}-{round(f_amp[1],2)} Hz for {sub}, {ses}, {str(ch)}:{'-'.join(logchan)}")    
                    
                    # d. Check if channel needs to be inverted for detection
                    if type(invert) == type(DataFrame()):
                        inversion = read_inversion(sub, ses, invert, ch, logger)
                        if not inversion:
                            inversion = infer_polarity(dset, annot, ch, 
                                                       chanset[ch], cat, None, 
                                                       self.stage, cycle, logger)
                    elif type(invert) == bool:
                        inversion = invert
                    else:
                        logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
                        logger.info('Check documentation for how to set up staging data:')
                        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                        logger.info('-' * 10)
                        return
    
                    # 5.a. Fetch data
                    segments = fetch(dset, annot, cat = cat, stage = self.stage,  
                                     cycle = cycle)

                    if len(segments)==0:
                        logger.warning(f"No valid data found for {sub}, {ses}, {self.stage}, Cycles:{cycle}.")
                        flag +=1
                        break
                    
                    # 5.b. Read data
                    if filter_opts['laplacian']:
                        try:
                            segments.read_data(filter_opts['lapchan'], chanset[ch]) 
                            laplace_flag = True
                        except:
                            logger.error(f"Channels listed in filter_opts['lapchan']: {filter_opts['lapchan']} are not found in recording for {sub}, {ses}.")
                            logger.warning("Laplacian filtering will NOT be run for {sub}, {ses}, {ch}. Check parameters under: filter_opts['lapchan']")
                            segments.read_data(ch, chanset[ch])
                            laplace_flag = False
                            flag += 1
                    else:
                        segments.read_data(ch, chanset[ch])
                    
                    
                    # 6. Define PAC object
                    pac = Pac(idpac = idpac, f_pha = f_pha, f_amp = f_amp, 
                              dcomplex = filter_opts['dcomplex'], 
                              cycle = filter_opts['filtcycle'], 
                              width = filter_opts['width'], 
                              n_bins = nbins,
                              verbose='ERROR')
                    
                    # 7.a. Divide segments based on concatenation
                    nsegs=[]
                    if model == 'whole_night':
                        nsegs = [s for s in segments]
                        seg_label = ['whole_night']
                    elif model == 'stage*cycle':
                        seg_label = []
                        for st in self.stage:
                            for cy in cycle_idx:
                                segs = [s for s in segments if st in s['stage'] if cy in s['cycle']][0]
                                nsegs.append(segs)
                                seg_label.append(f'{st}_cycle{cy}')
                    elif model == 'per_cycle':
                        seg_label = []
                        for cy in cycle_idx:
                            segs = [s for s in segments if cy in s['cycle']][0]
                            nsegs.append(segs)
                            seg_label.append(f'cycle{cy}')
                    elif model == 'per_stage':
                        seg_label = []
                        for st in self.stage:
                            segs = [s for s in segments if st in s['stage']][0]
                            nsegs.append(segs)
                            seg_label.append(f'{st}')
                    
                    # 7.b. Loop over segments
                    logger.info('')
                    for nsg in range(len(nsegs)):
                        seg = nsegs[nsg]
                        logger.debug(f'Analysing {seg_label[nsg]}')
                        
                        # 7.c. Create blocks
                        out = dict(seg)
                        data = seg['data']
                        timeline = data.axis['time'][0]
                        out['start'] = timeline[0]
                        out['end'] = timeline[-1]
                        out['duration'] = len(timeline) / data.s_freq

                        epoch = int(ceil((50/mean(frequency_phase))*data.s_freq)) # epoch length of 50 phase cycles
                        ms = int(ceil( len(timeline) / epoch )) # number of epochs in data
                        longamp = zeros((ms, epoch)) # initialise (blocked) amplitude series
                        longpha = zeros((ms, epoch)) # initialise (blocked) phase series 
                        ampbin = zeros((ms, nbins)) # initialise mean
                        
                        
                        
                        # 7.d Apply filtering (if required)
                        if filter_opts['laplacian']:
                            selectchans = filter_opts['lapchan']
                        else:
                            selectchans = ch
                        
                        # 7.d. Notch filters
                        filtflag = 0
                        if filter_opts['notch']:
                            data.data[0], filtflag = notch_mne(data, oREF=filter_opts['oREF'], 
                                                        channel=selectchans, 
                                                        freq=filter_opts['notch_freq'],
                                                        rename=filter_opts['laplacian_rename'],
                                                        renames=filter_opts['renames'],
                                                        montage=filter_opts['montage'])
                            
                        if filter_opts['notch_harmonics']: 
                            data.data[0], filtflag = notch_mne2(data, oREF=filter_opts['oREF'], 
                                                      channel=selectchans, 
                                                      rename=filter_opts['laplacian_rename'],
                                                      renames=filter_opts['renames'],
                                                      montage=filter_opts['montage'])    
                        
                        # 7.e. Bandpass filters
                        if filter_opts['bandpass']:
                            data.data[0], filtflag = bandpass_mne(data, oREF=filter_opts['oREF'], 
                                                      channel=selectchans,
                                                      highpass=filter_opts['highpass'], 
                                                      lowpass=filter_opts['lowpass'], 
                                                      rename=filter_opts['laplacian_rename'],
                                                      renames=filter_opts['renames'],
                                                      montage=filter_opts['montage'])
                        
                        # 7.f. Laplacian transform
                        if filter_opts['laplacian'] and laplace_flag:
                            data.data[0], filtflag = laplacian_mne(data, 
                                                 filter_opts['oREF'], 
                                                 channel=selectchans, 
                                                 ref_chan=chanset[ch], 
                                                 laplacian_rename=filter_opts['laplacian_rename'], 
                                                 renames=filter_opts['renames'],
                                                 montage=filter_opts['montage'])
                            data.axis['chan'][0] = asarray([x for x in chanset])
                            selectchans = ch
                            dat = data[0]
                        else:
                            dat = data()[0][0]
                            
                        # If any errors occured during filtering, break
                        if filtflag > 0:
                            break
                            
                        # 7.g. Fix polarity of recording
                        if inversion:
                            dat = dat*-1 
                        
                        for s in range(ms):
                            
                            datseg = dat[s*epoch:(s+1)*epoch]
                            
                            #  pad incomplete final block with nan
                            if len(datseg) < epoch:
                                datseg = concatenate((datseg, empty(epoch-len(datseg))))
                            
                            # Print out progress
                            if progress:
                                j = s/ms
                                sys.stdout.write('\r')
                                sys.stdout.write(f"                      Progress: [{'»' * int(50 * j):{50}s}] {int(100 * j)}%")
                                sys.stdout.flush()
                            
                            # 7.h. Obtain phase signal
                            pha = squeeze(pac.filter(data.s_freq, datseg, ftype='phase'))
                            
                            if len(pha.shape)>2:
                                pha = squeeze(pha)
                            
                            # 7.i. obtain amplitude signal
                            amp = squeeze(pac.filter(data.s_freq, datseg, ftype='amplitude'))
                            if len(amp.shape)>2:
                                amp = squeeze(amp)
                            
                            # 7.j.  put data in blocks (for surrogate testing)
                            longpha[s] = pha
                            longamp[s] = amp
                            
                            # 7.k. put averaged data in list (for preferred phase)
                            ampbin[s, :] = mean_amp(pha, amp, nbins=nbins)
        
                        # 8.a. if number of events not divisible by block length,
                        #    pad incomplete final block with randomly resampled events
                        sys.stdout.write('\r')
                        sys.stdout.flush()
                        
                        # b. Calculate Coupling Strength
                        mi = zeros((longamp.shape[0],1))
                        mi_pv = zeros((longamp.shape[0],1))
                        for r,row in enumerate(longamp): 
                            amp = row
                            pha = longpha[r]
                            amp = amp[~isnan(amp)]
                            if len(amp) < 1:
                                mi[r] = nan
                                mi_pv[r] = nan
                            else:
                                pha = reshape(pha,(1,1,len(pha)))
                                amp = reshape(amp,(1,1,len(amp)))
                                mi[r] = pac.fit(pha, amp, n_perm=400,random_state=5,
                                              verbose=False)[0][0]
                                mi_pv[r] = pac.infer_pvalues(p=0.95, mcp='fdr')[0][0]
    
                        ## c. Calculate preferred phase
                        ampbin = ampbin / ampbin.sum(-1, keepdims=True) # normalise amplitude
                        ampbin = ampbin.squeeze()
                        ampbin = ampbin[~isnan(ampbin[:,0]),:] # remove nan trials
                        ab = ampbin
                        
                        # d. Create bins for preferred phase
                        vecbin = zeros(nbins)
                        width = 2 * pi / nbins
                        for n in range(nbins):
                            vecbin[n] = n * width + width / 2  
                        
                        # e. Calculate mean direction (theta) & mean vector length (rad)
                        ab_pk = argmax(ab,axis=1)
                        theta = circ_mean(vecbin,histogram(ab_pk,bins=nbins, 
                                                            range=(0,nbins))[0])
                        theta_deg = degrees(theta)
                        if theta_deg < 0:
                            theta_deg += 360
                        rad = circ_r(vecbin, histogram(ab_pk,bins=nbins)[0], d=width)
                        
                        # f. Take mean across all segments/events
                        ma = nanmean(ab, axis=0)
                        
                        # g. Correlation between mean amplitudes and phase-giving sine wave
                        sine = sin(linspace(-pi, pi, nbins))
                        sine = interp(sine, (sine.min(), sine.max()), (ma.min(), ma.max()))
                        rho, pv1 = circ_corrcc(ma, sine)
    
                        # h. Rayleigh test for non-uniformity of circular data
                        ppha = vecbin[ab.argmax(axis=-1)]
                        z, pv2 = circ_rayleigh(ppha)
                        pv2 = round(pv2,5)
                        
                        # 9.a Export and save data
                        if adap_bands_phase == 'Fixed':
                            phadap = '-fixed'
                        else:
                            phadap = '-adap'
                        if adap_bands_amplitude == 'Fixed':
                            ampadap = '-fixed'
                        else:
                            ampadap = '-adap'    
                        freqs = f'pha-{f_pha[0]}-{f_pha[1]}Hz{phadap}_amp-{f_amp[0]}-{f_amp[1]}Hz{ampadap}'
                        if model == 'whole_night':
                            stagename = '-'.join(self.stage)
                            outputfile = '{}/{}_{}_{}_{}_{}_pac_parameters.csv'.format(
                                            outpath,sub,ses,fnamechan,stagename,freqs)
                        elif model == 'stage*cycle':    
                            outputfile = '{}/{}_{}_{}_{}_cycle{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,self.stage[nsg],cycle_idx[nsg],freqs)
                        elif model == 'per_stage':
                            outputfile = '{}/{}_{}_{}_{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,self.stage[nsg],freqs)
                        elif model == 'per_cycle':
                            stagename = '-'.join(self.stage)
                            outputfile = '{}/{}_{}_{}_{}_cycle{}_{}_pac_parameters.csv'.format(
                                          outpath,sub,ses,fnamechan,stagename,cycle_idx[nsg],freqs)
    
                        # b. Save cfc metrics to dataframe
                        d = DataFrame([nanmean(pac.pac), nanmean(mi), nanmedian(mi_pv), theta, 
                                        theta_deg, rad, rho, pv1, z, pv2])
                        d = d.transpose()
                        d.columns = ['mi_raw','mi_norm','pval1','pp_radians','ppdegrees','mvl',
                                      'rho','pval2','rayl','pval3']
                        d.to_csv(path_or_buf=outputfile, sep=',')
                        
                        # c. Save binned amplitudes to pickle file
                        outputfile = outputfile.split('_pac_parameters.csv')[0] + '_mean_amps'
                        save(outputfile, ab)
       
        ### 10. Check completion status and print
        if flag == 0:
            logger.debug('Phase-amplitude coupling finished without ERROR.')  
        else:
            logger.warning('Phase-amplitude coupling finished with WARNINGS. See log for details.')
        
        #self.tracking = tracking   ## TO UPDATE - FIX TRACKING
        
        return                                     




            

            





from .cfc_func import circ_wwtest, circ_kappa, mean_amp, klentropy, cohend
from .plots import plot_mean_amps, plot_prefphase, plot_prefphase_group, plot_meanamps_group
from .synchrony import event_sync, event_sync_dataset

__all__ = ['circ_wwtest', 'circ_kappa', 'mean_amp', 'klentropy', 'cohend',
           'plot_mean_amps', 'plot_prefphase', 
           'plot_prefphase_group', 'plot_meanamps_group',
           'event_sync', 'event_sync_dataset']




"""Packages containing power spectral analyses

"""
from .psa import (Spectrum, default_epoch_opts, default_event_opts, 
                  default_fooof_opts, default_filter_opts, default_frequency_opts, 
                  default_general_opts)
from .bandpower import (BandPowerTimecourse, default_bandpower_opts)
from .spectrogram import event_spectrogram, event_spectrogram_grouplevel



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  8 15:10:35 2024

@author: nathancross
"""

import logging
import sys



class CustomFormatter(logging.Formatter):
    
    cyan = "\x1b[36;20m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[1;33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format1 = "%(message)s"
    format2 = "%(asctime)s - %(name)s - %(message)s "
    format3 = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    format4 = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    # Check if colors are supported (i.e., output is a terminal)
    use_colors = sys.stdout.isatty()

    # Define format maps
    FORMATS_COLORED = {
        logging.INFO: cyan + format1 + reset,
        logging.DEBUG: grey + format2 + reset,
        logging.WARNING: yellow + format3 + reset,
        logging.ERROR: bold_red + format4 + reset,
        logging.CRITICAL: bold_red + format3 + reset
    }

    FORMATS_PLAIN = {
        logging.INFO: format1,
        logging.DEBUG: format2,
        logging.WARNING: format3,
        logging.ERROR: format4,
        logging.CRITICAL: format3
    }

    def format(self, record):
        log_fmt = (self.FORMATS_COLORED.get(record.levelno) if self.use_colors 
                   else self.FORMATS_PLAIN.get(record.levelno))
        formatter = logging.Formatter(log_fmt,"%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

class CustomLogfileFormatter(logging.Formatter):
    
    format1 = "%(message)s"
    format2 = "%(asctime)s - %(name)s - %(message)s "
    format3 = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    format4 = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.INFO: format1,
        logging.DEBUG: format2,
        logging.WARNING: format3,
        logging.ERROR: format4,
        logging.CRITICAL: format3
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt,"%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

def create_logger(name):
    logger = logging.getLogger(name)
    logger.handlers.clear()
    logger.setLevel(logging.DEBUG)
    
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    
    ch.setFormatter(CustomFormatter())
    
    logger.addHandler(ch)
    logger.propagate = False
    return logger

def create_logger_empty():
    
    # Set logging 
    logging.basicConfig(level=logging.DEBUG,
                        format="",
                        handlers=[logging.StreamHandler(sys.stdout)])
    logger = logging.getLogger()
    logger.handlers.clear()
    logger.propagate = False
    
    return logger

def create_logger_basic():
    
    # Set logging 
    logging.basicConfig(level=logging.DEBUG,
                        format="%(message)s",
                        handlers=[logging.StreamHandler(sys.stdout)])
    logger = logging.getLogger()
    logger.handlers.clear()
    logger.propagate = False
    
    return logger

def create_logger_outfile(logfile, name=None):

    # logging.basicConfig(level=logging.DEBUG,
    #                     format="%(message)s", 
    #                     handlers=[logging.StreamHandler(sys.stdout)])
    if name:
        logger = logging.getLogger(name)
    else:
        logger = logging.getLogger()
    
    logger.handlers.clear()
    
    # Logger to print to outfile
    file_log_handler = logging.FileHandler(f'{logfile}', mode='a')
    file_log_handler.setLevel(logging.DEBUG)
    file_log_handler.setFormatter(CustomLogfileFormatter())
    logger.addHandler(file_log_handler)
    
    # Logger to print to console
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)
    
    logger.setLevel(logging.DEBUG)

    logger.propagate = False

    return logger




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 15 11:33:16 2022

@author: nathancross
"""
from .logs import create_logger
import mne
import numpy as np
from pandas import DataFrame
import copy
from operator import itemgetter
from os import listdir, path, mkdir
from datetime import datetime
from wonambi import Dataset, graphoelement
from wonambi.attr.annotations import Annotations, create_empty_annotations
from wonambi.trans import fetch
from wonambi.detect.spindle import transform_signal
from scipy.signal import butter, filtfilt, find_peaks,  periodogram
import shutil
from .logs import create_logger
from .load import check_adap_bands
from .audit import check_fooof

                
def remove_event(annot, evt_name, chan = None, stage = None):
    
    ''' Remove event from annotation
        Workaround function because wonambi.attr.annotations.remove_event()
        is not working properly (and I couldn't figure out why).
    '''

    if not chan and not stage:
        evts = [x for x in annot.get_events(name=evt_name)]
    elif not stage:
        evts = [x for x in annot.get_events(name=evt_name) 
                if chan in x['chan']]
    elif not chan:
        evts = [x for x in annot.get_events(name=evt_name) 
                if stage in x['stage']]
    else:
        evts = [x for x in annot.get_events(name=evt_name) 
                if stage in x['stage'] if chan in x['chan']]
    
    annot.remove_event_type(name=evt_name)
    grapho = graphoelement.Graphoelement()
    grapho.events = evts       
    grapho.to_annot(annot, evt_name)
    
    return


def remove_duplicate_evts(annot, evt_name, chan = None, stage = None):
    
    ''' Removes any events in annotations that are duplicate.
        Workaround function because wonambi.attr.annotations.remove_event()
        is not working properly (and I couldn't figure out why).
    '''
    evts = annot.get_events(name = evt_name, chan = chan, stage = stage)
    evts_trim = copy.deepcopy(evts)
    for e, event in enumerate(evts[:-1]):
        starttime = event['start']
        i = 0
        for ee, eevent in enumerate(evts_trim):
                if eevent['start'] == starttime:
                    if i == 0:
                        None
                    elif i >0:
                        del(evts_trim[ee])
                    i=+1
    
    if chan:
        evts = [x for x in annot.get_events(name = evt_name) if chan not in x['chan']]
    else:
        evts = []
    
    annot.remove_event_type(name = evt_name)
    grapho = graphoelement.Graphoelement()
    grapho.events = evts + evts_trim          
    grapho.to_annot(annot, evt_name)
    
    return


def merge_events(annot, evt_name, chan = None, stage = None, segments = None):
    
    ''' Merges any events in annotations that are overlapping in time.
        # NOTE: This merges events with the SAME NAME ONLY. To merge 2 or more
                event types, use the function merge_2_events()
    '''
    
    evts = annot.get_events(name=evt_name, chan = chan, stage = stage)
    merged_events = []
    
    removed = []
    for e, event in enumerate(evts):
        if e not in removed:
            evts_trim = [item for i, item in enumerate(evts) if i != e]
        
            matches = ( [x for x in evts_trim if 
                         x['start'] <= event['start'] <= x['end']] +
                        [x for x in evts_trim if 
                         x['start'] <= event['end'] <= x['end']]
                      )
            
            #Find start and end times of merged events
            if len(matches) > 0:
                event['start'] = min([x['start'] for x in matches])
                event['end'] = max([x['end'] for x in matches])
                
                # Save to new event list
                merged_events.append(event)
            
            # Flag that these events have already been merged and removed
            for m in matches:
                index = [i for i, evt in enumerate(evts) if
                         evt['start'] == m['start'] if evt['end'] == m['end']]
                removed.append(index)
    
    events_trim = [evts[i] for i, x in enumerate (evts) if not i in removed]
    merged_events = merged_events + events_trim
    
    
    if chan:
        evts = [x for x in annot.get_events(name = evt_name) if chan not in x['chan']]
    else:
        evts = []
    
    annot.remove_event_type(name = evt_name)
    grapho = graphoelement.Graphoelement()
    grapho.events = evts + merged_events         
    grapho.to_annot(annot, evt_name)       


    return    

def merge_2_events(annot, events, new_event_name, chan = None, stage = None, 
                   segments = None):
    
    ''' Merges 2 different event types in annotations (irregardless of any 
            overlap). However, this will then also combine any events that 
            have a temporal overlap into the 1 event.
    '''
    # Get each event from annotations file
    evts = []
    for evt_name in events:
        evts += annot.get_events(name = evt_name, chan = chan, stage = stage)

    # Rename both events into a single event name
    evts = [{**d, "name": new_event_name} if "name" in d else d for d in evts]
    
    # Save new events to Annotations file
    grapho = graphoelement.Graphoelement()
    grapho.events = evts         
    grapho.to_annot(annot, new_event_name)       
    
    # Finally, combine any overlapping events
    merge_events(annot, new_event_name, chan, stage, segments)

    return    


def remove_duplicate_evts_bids(in_dir, out_dir, chan, grp_name, rater, 
                               cat=(0, 0, 0, 0), stage = None, 
                               evt_name = None, part = 'all', visit = 'all', 
                               param_keys = None):

    if path.exists(in_dir):
            print(in_dir + " already exists")
    else:
        mkdir(in_dir)
    
    print('')
    print(f'Time start: {datetime.now()}')
    print(f'Removing duplicates from files in directory {in_dir}')  
    print(f'Event = {evt_name[0]}')
    print('')
    
    if evt_name == None:
        evt_name = ['spindle']
    
    # Loop through subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(in_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    part.sort()                
    for i, p in enumerate(part):
        # Loop through visits
        if visit == 'all':
            visit = listdir(in_dir + '/' + p)
            visit = [x for x in visit if not '.' in x]
        
        print(f'Removing duplicate events for Subject {p}..')
        
        visit.sort()
        for v, vis in enumerate(visit): 
            if not path.exists(out_dir + '/' + p + '/' + vis + '/'):
                mkdir((out_dir + '/' + p + '/' + vis + '/'))
            backup = out_dir + '/' + p + '/' + vis + '/' #backup folder

            # Define files
            rec_dir = in_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(rec_dir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_file) == 0:                
                print(f'WARNING: whale_it has not been run for Subject {p}, skipping..')
                
            else:
                # backup file
                backup_file = (backup + xml_file[0])
                shutil.copy(rec_dir + xml_file[0], backup_file) 
                
                # Import Annotations file
                annot = Annotations(backup_file, 
                                    rater_name=rater)
                
                # Run through channels
                for ch, channel in enumerate(chan):
                    chan_ful = channel + ' (' + grp_name + ')'
                                    
                    ### WHOLE NIGHT ###
                    # Select and read data
                    print('Reading data for ' + p + ', visit ' + vis + ' ' + channel)
                    
                    # Retrieve events
                    remove_duplicate_evts(annot, evt_name, chan_ful, stage)

    return


def clean_annots(xml_dir, out_dir, rater, keep_evts = None, subs = 'all', 
                 sessions = 'all', logger = create_logger('Clean annotations')):
    
    '''
        Copies the annotations files from inside xml_dir into out_dir and 
        removes all annotations from the files, keeping only staging info.
        
    '''
    
    logger.debug('')
    logger.debug(f'Time start: {datetime.now()}')
    logger.debug(f'Removing events from files in directory {xml_dir}')  
    logger.debug(f'Keeping events = {keep_evts}')
    logger.debug(f'Saving new annotations files to directory {out_dir}')
    logger.debug('')
    
    # Loop through subjects
    if isinstance(subs, list):
        None
    elif subs == 'all':
            subs = listdir(xml_dir)
            subs = [x for x in subs if not '.' in x]
    else:
        logger.error("'subs' must either be an array of subject ids or = 'all' **CHECK**")
    
    subs.sort()                
    for i, sub in enumerate(subs):
        # Loop through visits
        if sessions == 'all':
            sessions = listdir(xml_dir + '/' + '/' + sub)
            sessions = [x for x in sessions if not '.' in x]
        sessions.sort()
        for v, ses in enumerate(sessions): 
            
            # Define files
            xdir = xml_dir + '/' + sub + '/' + ses + '/'
            xml_files = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_files) == 0:                
                logger.warning(f'No xml files found for Subject {sub}, skipping..')
  
            else:
                for xml_file in xml_files:
                    ## Copy annotations file before beginning
                    if not path.exists(out_dir):
                        mkdir(out_dir)
                    if not path.exists(out_dir + sub):
                        mkdir(out_dir + sub)
                    if not path.exists(out_dir + sub + '/' + ses):
                        mkdir(out_dir + sub + '/' + ses)
                    backup = out_dir + sub + '/' + ses + '/'
                    backup_file = (f'{backup}{xml_file[0]}')
                    shutil.copy(xdir + xml_file, backup_file)
                                
                    # Import Annotations file
                    annot = Annotations(backup_file, rater_name=rater)
                    evts = list(set([x['name'] for x in annot.get_events()]))
                    
                    logger.debug(f'Removing events for {sub}, {ses}..')
                    if keep_evts == None:
                        keep_evts = []
                    for ev in evts:
                        if ev not in keep_evts:
                            annot.remove_event_type(name=ev)


def merge_epochs(epochs):
    
    ''' Function to merge epochs of the same stage together into larger segments,
        to extract the start and end times of periods of sleep stages. An 
        exception is granted for single epochs (30s) splitting a larger segment,
        in this case the single epoch is 'absorbed' by the larger segment.
        
        Parameters
        ----------
        epochs  - A list of dict corresponding to segments of data. Each dict 
                  must contain the following parameters: 'start', 'end', 'stage'.
        
        Returns
        -------
        A list of dict corresponding to segments of data. Each dict contains the 
        following parameters: 'start', 'end', 'stage'.
        
    '''
    

    merged = []
    current = epochs[0].copy()
    skip_next = False  # Flag to track if we are skipping an interrupting element

    for i in range(1, len(epochs)):
        entry = epochs[i]

        if skip_next:
            skip_next = False
            continue

        if entry["stage"] == current["stage"] and entry["start"] <= current["end"]:
            # Merge as usual
            current["end"] = max(current["end"], entry["end"])
        elif (
            i + 1 < len(epochs) and epochs[i + 1]["stage"] == current["stage"]
        ):  
            # If the next element matches the current stage, skip this one
            skip_next = True
            current["end"] = max(current["end"], epochs[i + 1]["end"])
        else:
            # Save the merged segment and move to the next one
            merged.append(current)
            current = entry.copy()

    # Append the last segment
    merged.append(current)
    return merged

                            

def merge_xmls(in_dirs, out_dir, chan, grp_name, stage = None, evt_name = None, 
               part = 'all', visit = 'all'):

    for pth in in_dirs:
        if not path.exists(pth):
            print(pth + " doesn't exist. Check in_dirs is set correctly.")
            
    else: 
        main_dir = in_dirs[0]
        sec_dir = in_dirs[1]
        
        print('')
        print(f'Time start: {datetime.now()}')
        print(f'Initiating the merging of xml files in directory {main_dir}...') 
        print(f'with the xml files in directory {sec_dir}...') 
        print('')
        
        # Get list of subjects
        if isinstance(part, list):
            None
        elif part == 'all':
                part = listdir(main_dir)
                part = [ p for p in part if not '.' in p]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
        
        
        # Loop through subjects  
        part.sort()              
        for i, p in enumerate(part):
            print(f'Merging xml files for Subject {p}..')
            # Get visits
            if visit == 'all':
                visit = listdir(main_dir + '/' + p)
                visit = [x for x in visit if not '.' in x]
            
            # Loop through visits
            visit.sort()
            for v, vis in enumerate(visit): 
                if not path.exists(main_dir + '/' + p + '/' + vis):
                    print(main_dir + '/' + p + '/' + vis + " doesn't exist. Check data is in BIDS format.")
                
                # Check for output directory and create
                if not path.exists(out_dir + '/'):
                     mkdir(out_dir + '/')
                if not path.exists(out_dir + '/' + p):
                    mkdir(out_dir + '/' + p)
                if not path.exists(out_dir + '/' + p + '/' + vis):
                   mkdir(out_dir + '/' + p + '/' + vis)
                output = out_dir + '/' + p + '/' + vis + '/' #backup folder
    
    
                # Define files
                if main_dir == sec_dir:
                
                    rec_dir = main_dir + '/' + p + '/' + vis + '/'
                    xml_file = [x for x in listdir(rec_dir) if x.endswith('.xml') 
                                if not x.startswith('.')] 
                    xml_file.sort()
                    
                else:
                    rec_dir = main_dir + '/' + p + '/' + vis + '/'
                    xml_file1 = [x for x in listdir(rec_dir) if x.endswith('.xml') 
                                if not x.startswith('.')]
                    rec_dir2 = sec_dir + '/' + p + '/' + vis + '/'
                    xml_file2 = [x for x in listdir(rec_dir2) if x.endswith('.xml') 
                                if not x.startswith('.')]
                    xml_file = xml_file1 + xml_file2
                    
                    
                # Merge files
                if len(xml_file) < 2:                
                    print(f'WARNING: Only 1 xml file found for Subject {p}, skipping..')
                else:
                    # Copy first file to output directory
                    output_file = (output + p + '-merged.xml')
                    shutil.copy(rec_dir + xml_file[0], output_file) 
                    
                    # Import Annotations file
                    annot = Annotations(output_file, 
                                        rater_name=None)
                    
                    for x,xml in enumerate(xml_file[1:]):
                        annot2 = Annotations(rec_dir + xml, 
                                        rater_name=None)
                    
                        # Run through channels
                        for ch, channel in enumerate(chan):
                            chan_ful = [channel + ' (' + grp_name + ')']
                                            
                            print('Merging on channel: {channel}')
                            
                            # Calculate event density (whole night)
                            if evt_name is not None:
                                for e,evt in enumerate(evt_name):
                                    evts = annot2.get_events(name=evt, chan = chan_ful, 
                                                             stage = stage)
                                    annot.add_events(evts)
                            else:
                                evts = annot2.get_events(name=None, chan = chan_ful, 
                                                         stage = stage)
                                annot.add_events(evts)

    return


def rainbow_merge_evts(xml_dir, out_dir, chan, grp_name, rater, segments = None, 
                       events = None, evt_name = None, part = 'all', 
                       visit = 'all'):
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    if evt_name == None:
        evt_name = ['spindle']
    
    # Loop through subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    part.sort()                
    for i, p in enumerate(part):
        # Loop through visits
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        print(" ")      
        print(f'Merging rainbows for Subject {p}..')
        print(r""" 
                 .##@@&&&@@##.
              ,##@&::%&&%%::&@##.
             #@&:%%000000000%%:&@#
           #@&:%00'         '00%:&@#
          #@&:%0'             '0%:&@#
         #@&:%0                 0%:&@#
        #@&:%0                   0%:&@#
        #@&:%0                   0%:&@#
                """,flush=True)   
        
        
        visit.sort()
        for v, vis in enumerate(visit): 
            if not path.exists(out_dir + '/' + p + '/' + vis + r'/'):
                mkdir(out_dir + '/' + p + '/' + vis + r'/')
            backup = out_dir + '/' + p + '/' + vis + r'/' #backup folder

            # Define files
            xdir = xml_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_file) == 0:                
                print(f'WARNING: no xml files found for Subject {p}, skipping..')
            elif len(xml_file) > 1:
                print(f'WARNING: multiple xml files found for Subject {p}, visit {v}. Check this! Skipping..')
            else:
                # backup file
                backup_file = (backup + p + '_trunc.xml')
                shutil.copy(xdir + xml_file[0], backup_file) 
                
                # Import Annotations file
                annot = Annotations(backup_file,rater_name=rater)
                if segments is not None:
                    times = []
                    for seg in segments:
                        
                        starts = [x['end'] for x in sorted(annot.get_events(name=seg[0]), 
                                                           key=itemgetter('start')) ]
                        ends = [x['start'] for x in sorted(annot.get_events(name=seg[1]), 
                                                           key=itemgetter('start')) ]
                        for i in range(0,len(starts)):
                            times.append((starts[i],ends[i]))
                else:
                    times = [(0,annot.last_second)]
                
                evts = []
                for t,time in enumerate(times):
                    
                    if events is not None:
                        for e in events:
                            evts.extend(annot.get_events(name=e,time=time)[:])
                    else:
                        evts = annot.get_events(time=time)[:]

            
                names = set([e['name'] for e in evts])  
                newevents = copy.deepcopy(evts)
                d1 = {'name': evt_name[0]}
                [x.update(d1) for x in newevents]
                

                for t,typ in enumerate(names):    
                    annot.remove_event_type(name=typ)
                
                grapho = graphoelement.Graphoelement()
                grapho.events = newevents          
                grapho.to_annot(annot)


    return


def rename_evts(xml_dir, out_dir, part, visit, evts):
    
    '''
       Renames events inside annotations files. The xmls are first copied to out_dir
       and the events are deleted from the copied xmls.
       Inputs are :
           xml_dir, out_dir = input xml_directory and output directory for new xmls
           evts             = dictionary for how events are to be renamed:
                               e.g. evts = {'Moelle2011':'spindle',
                                            'Lacourse2018':'spindle,
                                            'Staresina2015':'so'}
    '''
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Get list of subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    # Loop through subjects  
    part.sort()              
    for i, p in enumerate(part):
        
        # Get visits
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not '.' in x]
        
        # Loop through visits
        visit.sort()
        for v, vis in enumerate(visit): 
            
            print(f'Subject {p}, visit {vis}..')
            
            # Check for input data
            if not path.exists(xml_dir + '/' + p + '/' + vis):
                print(xml_dir + '/' + p + '/' + vis + " doesn't exist. Check data is in BIDS format.")
            
            # Check for output directory and create
            if not path.exists(out_dir + '/'):
                 mkdir(out_dir + '/')
            if not path.exists(out_dir + '/' + p):
                mkdir(out_dir + '/' + p)
            if not path.exists(out_dir + '/' + p + '/' + vis):
               mkdir(out_dir + '/' + p + '/' + vis)
            output = out_dir + '/' + p + '/' + vis + '/' #backup folder

            # Define files
            xdir = xml_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            xml_file.sort()
            
            if len(xml_file) > 1:                
                print(f'WARNING: Multiple xml files for Subject {p}, visit {vis} - check this - skipping..')
                
            else:
                
                # Copy first file to output directory
                output_file = (output + xml_file[0])
                shutil.copy(xdir + xml_file[0], output_file) 
                
                # Import Annotations file
                annot = Annotations(output_file, rater_name=None)
                
                # Rename events
                for e,event in enumerate(evts):
                    old = event
                    new = evts[event]
                    
                    print(f"Renaming event '{old}' to '{new}'")
                    
                    annot.rename_event_type(old, new)
            
                
    return
                

def laplacian_mne(data, channel, ref_chan, oREF = None, laplacian_rename=False, 
                  renames=None, montage='standard_alphabetic', 
                  logger = create_logger('Laplacian filter')):
    
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if oREF:
        try:
            dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        except Exception as error:
            logger.error(error)
            logger.error("This is due to an incorrect naming of an online "
                         "reference 'oREF' (see user guide):")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            return data, 1
    
    if laplacian_rename:
        dig.rename_channels(renames, allow_duplicates=False)
    
    info = mne.create_info(ch_names, data.s_freq,verbose=None)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    a.set_eeg_reference(ref_chan,verbose=40)
    raw_csd = mne.preprocessing.compute_current_source_density(a, verbose=40)
    data = raw_csd.get_data(picks=channel)
    
    return data, 0


def notch_mne(data, channel, freq, oREF = None, rename=False, renames=None, 
              montage='standard_alphabetic', 
              logger = create_logger('Notch filter')):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        try:
            dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        except Exception as error:
            logger.error(error)
            logger.error("This is due to an incorrect naming of an online "
                         "reference 'oREF' (see user guide):")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            return data, 1
    
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    anotch = a.notch_filter(freq,verbose=40)
    data = anotch.get_data(picks=channel)
    
    return data, 0 

def notch_mne2(data, channel, oREF = None, rename=False, renames=None,
               montage = 'standard_alphabetic', 
               logger = create_logger('Notch filter')):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        try:
            dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        except Exception as error:
            logger.error(error)
            logger.error("This is due to an incorrect naming of an online "
                         "reference 'oREF' (see user guide):")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            return data, 1
        
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    anotch = a.notch_filter(freqs=None, filter_length='auto', 
                            notch_widths=None, trans_bandwidth=1, 
                            method='spectrum_fit', verbose=None)
    data = anotch.get_data(picks=channel)
    
    return data, 0
    

def bandpass_mne(data, channel, highpass, lowpass, oREF = None, rename=False,
                 renames=None, montage='standard_alphabetic', 
                 logger = create_logger('Bandpass filter')):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        try:
            dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        except Exception as error:
            logger.error(error)
            logger.error("This is due to an incorrect naming of an online "
                         "reference 'oREF' (see user guide):")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            return data, 1
        
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    a_filt = a.filter(highpass, lowpass,verbose=40)
    data = a_filt.get_data(picks=channel)
    
    return data, 0

def csv_stage_import(edf_file, xml_file, hypno_file, rater):
    
    ''' This function creates a new annoations file and imports staging from a csv 
        file. The staging should be in the 'Alice' format, for further information 
        see: wonambi.attr.annotations.
    '''

    dataset = Dataset(edf_file)
    create_empty_annotations(xml_file, dataset)
    annot = Annotations(xml_file)
    annot.import_staging(hypno_file, 'alice', rater_name=rater, rec_start=dataset.header['start_time'])
    




def reconstruct_stitches(dat, stitches, s_freq, replacement = 0):
    orig_length = int(stitches[-1][1]*s_freq)  # Length of original array2
    
    # Step 1: Create a zero-filled array2
    dat_reconstructed = np.zeros(orig_length)
    dat_reconstructed.fill(replacement)
    
    # Step 2: Reinsert segments into dat_reconstructed
    idx = 0  # Index to track position in array1
    
    for start, end in stitches:
        start_idx = int(start * s_freq)
        end_idx = int(end * s_freq)
        segment_length = end_idx - start_idx
        
        # Insert the segment from array1
        dat_reconstructed[start_idx:end_idx] = dat[idx:idx + segment_length]
        
        # Update the array1 index tracker
        idx += segment_length
        
    return dat_reconstructed


def adap_bands_setup(self, adap_bands, frequency, subs, sessions, chan, ref_chan,
                     stage, cat, concat_cycle, cycle_idx, logger):
    
    if adap_bands not in ['Fixed', 'Auto', 'Manual']:
        logger.critical("The argument 'adap_bands' must be either one of"
                        f"'Fixed', 'Auto', 'Manual', not: '{adap_bands}'")
        flag = 'error'
   
    if adap_bands == 'Fixed':
        flag = 'None'
        if not frequency:
            frequency = (11,16)
    elif adap_bands == 'Manual':
        logger.debug(f"Checking for spectral peaks in {self.rootpath}/'tracking.tsv' ")
        flag = check_adap_bands(self.rootpath, subs, sessions, chan, logger)
        if flag == 'error':
            logger.critical('Spindle detection finished with ERRORS. See log for details.')
            
        elif flag == 'review':
            logger.info('')
            logger.warning("Some spectral peak entries in 'tracking.tsv' are "
                           "inconsistent or missing. In these cases, detection will "
                           f"revert to fixed bands: {frequency[0]}-{frequency[1]}Hz")
            logger.info('')
    elif adap_bands == 'Auto': 
        if not frequency:
            frequency = (9,16) 
            flag = 'None'
        self.track(subs, sessions, step = 'fooof', show = False, log = False)
        if not type(chan) == type(DataFrame()):
            logger.critical("For adap_bands = Auto, the argument 'chan' must "
                            "be 'None' and specfied in 'tracking.csv'")
            flag = 'error'
        else:
            flag, pk_chan, pk_sub, pk_ses = check_fooof(self, frequency, 
                                                              chan, ref_chan, 
                                                              stage, 
                                                              cat,
                                                              cycle_idx, 
                                                              logger)
            if flag == 'error':
                logger.critical('Error in reading channel names, check tracking sheet.')
                logger.info("Check documentation for how to set up channel names in tracking.tsv':")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
            elif flag == 'review':
                logger.debug("Spectral peaks have not been found for all subs, "
                             "analysing the spectral parameters prior to spindle detection..")
                for (sub,ses) in zip(pk_sub, pk_ses):
                    self.detect_spectral_peaks(subs = [sub], 
                                           sessions = [ses], 
                                           chan = pk_chan, 
                                           frequency = frequency,
                                           stage = stage, cycle_idx = cycle_idx,
                                           concat_cycle = concat_cycle, 
                                           concat_stage=True)    
    return frequency, flag


def infer_polarity(dset, annot, chan, ref_chan, cat = (1,1,1,1), evt_type = None, 
                   stage = None, cycle = None, logger = create_logger('Check polarity')):

    segments = fetch(dset, annot, cat, evt_type, stage, cycle)
    segments.read_data(chan, ref_chan)
    
    s_freq = dset.header['s_freq']
    
    X = segments.segments[0]['data'].data[0][0]
    
    XNormed = (X - X.mean())/(X.std())
    
    X_trans = transform_signal(XNormed, s_freq, 'double_sosbutter', 
                               method_opt={'freq':(0.1,4.5), 'order':3})
    
    ## Positive component of signal
    X_pos = copy.deepcopy(X_trans)
    X_pos[X_pos<0] = np.nan
    
    # Positive peaks
    peaks_pos = find_peaks(X_pos, height = np.nanstd(X_pos)) 
    pos_peaks_mean = np.nanmean(X_pos[peaks_pos[0]])
    
    # Positive mean
    pos_mean = np.nanmean(X_pos[X_pos>np.nanpercentile(X_pos, 99)])
    
    # Positive PSD
    result = flip_and_switch(X_pos)
    Sxx = periodogram(result, s_freq)
    pos_Sxx_mean = np.nanmean(Sxx[1][0:2000])
    
    ## Negative component of signal
    X_neg = copy.deepcopy(X_trans)
    X_neg[X_neg>0] = np.nan
    X_neg = X_neg * -1
    
    # Negative peaks
    peaks_neg = find_peaks(X_neg, height = np.nanstd(X_neg)) 
    neg_peaks_mean = np.nanmean(X_neg[peaks_neg[0]])
    
    # Negative mean
    neg_mean = np.nanmean(X_neg[X_neg>np.nanpercentile(X_neg, 99)])
    
    # Negative PSD
    result = flip_and_switch(X_neg)
    Sxx = periodogram(result, s_freq)
    neg_Sxx_mean = np.nanmean(Sxx[1][0:2000])
    
    # Compare Postive to Negative
    rating = ( 
              int(pos_peaks_mean>neg_peaks_mean) + 
              int(pos_Sxx_mean>neg_Sxx_mean) + 
              int(pos_mean > neg_mean) 
              )

    if rating > 0:
        logger.debug('Signal polarity appears to be reversed.')
        invert = True
    else:
        invert = False
        logger.debug('Signal polarity appears correct.')
    return invert


def flip_and_switch(data):
    
    # Step 1: Identify segments between NaNs
    segments = []
    current_segment = []
    
    for i, val in enumerate(data):
        if not np.isnan(val):
            current_segment.append(val)
        elif current_segment:
            segments.append(np.array(current_segment))
            current_segment = []
    
    # Add the last segment if it exists
    if current_segment:
        segments.append(np.array(current_segment))
    
    # Step 2: Flip every even-indexed segment (starting at index 1)
    for i in range(len(segments)):
        if i % 2 == 1:  # 0-based index: 1st, 3rd, etc.
            segments[i] = -segments[i]
    
    # Step 3: Stitch the segments together
    result = np.concatenate(segments)
    
    return result


def check_data_length(data, duration, logger = create_logger('Data check')):
    
    if any([data(trial=x).shape[1]<= duration*data.s_freq for 
            x in range(data.number_of('trial'))]):
        logger.warning('Length of data too short!!')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 18 19:52:50 2024

@author: ncro8394
"""



def eel():
    
    '''
        Excitatory Event Locater (EEL)
    
        An automative detection of cortical arousals.
    
    '''
    
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 18 16:29:30 2025

@author: ncro8394
"""


from numpy import (argsort, asarray, concatenate, cumsum, diff, exp, isnan, 
                   logical_and, mean, nan, nanmean, nanstd, sort, var, zeros)
from os import listdir, mkdir, path
from pandas import concat, DataFrame, read_csv
from pathlib import Path
from wonambi import Dataset
from wonambi.detect.spindle import transform_signal
from scipy.integrate import trapz
from scipy.signal import find_peaks, periodogram
from scipy.stats import variation
from sleepecg import detect_heartbeats
from .logs import create_logger
from ..events.sand import detect_ecg_artifacts





class SQUID:
    """
    SQUID: Signal Quality Inference and Detection

    This class orchestrates signal quality assessment over a BIDS-formatted dataset,
    using modular QC pipelines for EEG, EMG, and ECG recordings.

    Parameters
    ----------
    bids_root : str or Path
        Path to the root of the BIDS-formatted dataset.

    filetype : str, optional
        Modality type to analyze ('eeg', 'ecg', 'emg'), by default 'eeg'.

    subjects : list of str, optional
        List of subject IDs to include (e.g., ['01', '02']). If None, all subjects are processed.

    sessions : list of str, optional
        List of session IDs. If None, sessions are inferred automatically.

    logger : logging.Logger, optional
        Custom logger instance. If None, a default logger is created.
    """

    def __init__(self, bids_root, filetype = '.edf', subjects = None, 
                 sessions = None, logger = None):
        self.bids_root = Path(bids_root)
        
        self.datapath = Path(f'{self.bids_root}/sourcedata/')
        if not path.exists(self.datapath):
            self.datapath = Path(f'{self.bids_root}/DATA/')

        if not path.exists(f'{bids_root}/derivatives'):
            mkdir(f'{bids_root}/derivatives')
        if not path.exists(f'{bids_root}/derivatives/QC'):
            mkdir(f'{bids_root}/derivatives/QC') 
        self.outpath = Path(f'{bids_root}/derivatives/QC')
            
        self.filetype = filetype
        self.subjects = subjects
        self.sessions = sessions
        self.logger = create_logger('SQUID')
        self.results = []

    
    def process_all(self, filt = None, chantype = ['eeg', 'eog', 'emg', 'ecg']):
        """
        Iterates over all subjects and sessions to run quality control.
        """
        subjects = self.subjects or self._get_all_subjects()

        for subj in subjects:
            sessions = self.sessions or self._get_sessions(subj)

            for sess in sessions:
                try:
                    self.logger.debug(f"Processing {subj}, {sess}")
                    self.process_subject(subj, chantype, filt, sess)
                except Exception as e:
                    self.logger.warning(f"Failed to process {subj}, {sess}: {e}")
                    continue

    def process_subject(self, subject_id, chantype, filt=None, session_id=None):
        """
        Process a single subject/session: infer, choose best, and compute QC.

        Parameters
        ----------
        subject_id : str
        session_id : str or None
        chantype : list
        """
        
        dset = self._load_dataset(subject_id, session_id)
        outdir = self._set_outdir(subject_id, session_id)
    
        # Define function mappings
        candidate_funcs = {
            'eeg': get_candidate_eeg,
            'eog': get_candidate_eog,
            'emg': get_candidate_emg
        }
    
        squid_funcs = {
            'eeg': squid_eeg,
            'eog': squid_eeg,  # assuming same function as eeg
            'emg': squid_emg
        }
    
        for c in chantype:
            if c not in candidate_funcs or c not in squid_funcs:
                self.logger.warning(f"Unsupported chantype: {c}")
                continue
    
            chans = candidate_funcs[c](dset, filt, logger=self.logger)
            if not chans:
                continue
    
            results = []
            for chan in chans:
                try:
                    signal = dset.read_data(chan=[chan]).data[0][0]
                    s_freq = dset.header['s_freq']
                    qc_row = squid_funcs[c](signal, s_freq, name=chan)
                    qc_row['subject'] = subject_id
                    qc_row['session'] = session_id
                    results.append(qc_row)
                except Exception as e:
                    self.logger.warning(f"Failed to process channel {chan}: {e}")
    
            if results:
                savename = f'{subject_id}_{session_id}' if session_id else subject_id
                df = concat(results, ignore_index=True)
                cols_first = ['subject', 'session', 'channel', 
                              'quality_score', 'quality_label']
                cols_rest = [col for col in df.columns if col not in cols_first]
                df = df[cols_first + cols_rest]
                filepath = f'{outdir}/{savename}_qc_report_{c}.csv'
                df.to_csv(filepath, index=False)


            # TO DO:
            # if c == 'ecg':
            #     chans = infer_ecg(dset)
            #     if chans:
            #         best = choose_best_emg(chans, dset)
            #         for chan in best:
            #             signal = dset.read_data(chan=chan).data[0][0]  # Adapt as needed
            #             s_freq = dset.get_sampling_rate(chan=chan)
            #             qc_row = squid_emg(signal, s_freq, name=chan)
            #             qc_row['subject'] = subject_id
            #             qc_row['session'] = session_id
            #             self.results.append(qc_row)
                
            #     filepath = f'{outdir}/qc_report_{c}.csv'
            #     self.save_results(filepath)
            

    def _load_dataset(self, subject_id, session_id):
        """
        Load a subject/session dataset. 
        """
        
        dpath = f'{self.datapath}/{subject_id}/{session_id}/eeg'
        if not path.exists(dpath):
            self.logger.warning(f'Directory not found for {subject_id}, {session_id}.')
            return
        
        filename = [x for x in listdir(dpath) if self.filetype in x]
        
        if len(filename) == 0:
            self.logger.warning(f"No {self.filetype} files found for "
                                f"{subject_id}, {session_id}")

        dset = Dataset(f'{dpath}/{filename[0]}')
        
        return dset
    
    def _set_outdir(self, subject_id, session_id):
        """
        Sets the output directory to save the audit. 
        """
        outdir = f'{self.outpath}/{subject_id}/'
        if not path.exists(f'{outdir}'):
            mkdir(f'{outdir}')
        outdir = f'{self.outpath}/{subject_id}/{session_id}/'
        if not path.exists(f'{outdir}'):
            mkdir(f'{outdir}')
            
        return outdir

    def _get_all_subjects(self):
        """
        Infer all subject IDs from the BIDS root.
        """
        subj_dirs = sorted(self.datapath.glob("sub-*"))
        return [d.name.split("/")[-1] for d in subj_dirs]

    def _get_sessions(self, subject_id):
        """
        Infer session IDs from a subject folder, or return [''] if no sessions.
        """
        subj_path = self.datapath / f"{subject_id}"
        ses_dirs = sorted(subj_path.glob("ses-*"))
        if ses_dirs:
            return [d.name.split("/")[-1] for d in ses_dirs]
        else:
            return ['']



def get_candidate_eeg(dset, filt=None,logger=create_logger('Get Candidate EEG')):
    """Return a list of plausible EEG channel names based on known naming conventions.

    Parameters
    ----------
    dset : object
        The dataset object containing signal metadata.
    filt : list of str, optional
        List of string patterns to filter EEG channels explicitly.
    logger : Logger object, optional
        Logger for status and error messages.

    Returns
    -------
    list of str
        Candidate EEG channel names (may be empty if no matches found).
    """
    h = dset.header

    if filt:
        if not isinstance(filt, list):
            raise ValueError("`filt` must be a list")
        eeg = [x for x in h["chan_name"] if any(l in x for l in filt)]

    else:
        eeg = [x for x in h["chan_name"] if any(l in x for l in ['C1', 'C2', 
                                                                'C3', 'C4', 
                                                                'C5', 'C6', 
                                                                'Cz', 'CZ',
                                                                'F1', 'F2', 
                                                                'F3', 'F4', 
                                                                'F5', 'F6', 
                                                                'Fz', 'FZ',
                                                                'Pz', 'PZ',
                                                                'P3', 'P4',
                                                                'Fpz', 'Oz',
                                                                'O1', 'O2'])]

        if len(eeg) == 0:
            eeg = [x for x in h["chan_name"] if 'E' in x and not 
                   any(l in x for l in ['EMG', 'ECG', 'EOG', 'Chin'])]

    return eeg


def get_candidate_eog(dset, logger=create_logger('Get Candidate EOG')):
    """
    Identify potential EOG channels from the dataset header based on naming heuristics.

    Parameters
    ----------
    dset : object
        The dataset object with header information.
    logger : logging.Logger, optional
        Logger for debug messages.

    Returns
    -------
    list
        List of candidate EOG channel names (may be empty if no candidates found).
    """
    h = dset.header

    # First and most obvious EOG label
    eyes = [x for x in h['chan_name'] if 'EOG' in x]

    # Fallback: channels with 'E' but excluding non-EOG types
    if len(eyes) == 0:
        eyes = [x for x in h['chan_name'] if 'E' in x and 
                all(excl not in x for excl in ['EMG', 'ECG'])]

    return eyes


def get_candidate_emg(dset, logger=create_logger('Get Candidate EMG')):
    """
    Identify potential EMG channels from the dataset header based on naming heuristics.

    Parameters
    ----------
    dset : object
        The dataset object with header information.
    logger : logging.Logger, optional
        Logger for debug messages.

    Returns
    -------
    list
        List of candidate EMG channel names.
    """
    h = dset.header

    # Most common EMG indicators
    emg = [x for x in h['chan_name'] if any(l in x for l in ['EMG', 'Chin', 'CHIN'])]

    return emg


def infer_eeg(dset, filt=None, num=1, logger=create_logger('Infer EEG')):
    
    """Return top-N EEG channels based on quality scores using SQUID QC.

    Parameters
    ----------
    dset : object
        Dataset object with header and signal data.
    filt : list of str, optional
        Optional patterns to filter candidate EEGs.
    num : int, optional
        Number of top-quality EEG channels to return.
    logger : Logger, optional
        Logger for messages.

    Returns
    -------
    list of str
        EEG channel names ranked by quality (or None if no suitable channels).
    """
    eeg = get_candidate_eeg(dset, filt=filt, logger=logger)

    if len(eeg) == 0 or len(eeg) > 10:
        logger.error("Unable to determine EEG channels from recording. "
                     "EEG names must be explicitly specified.")
        logger.info("Check documentation for how to specify channels:")
        logger.info("https://seapipe.readthedocs.io/en/latest/index.html")
        logger.info("-" * 10)
        return None

    # Try to find ECG to pass into choose_best_eeg if needed
    ecg = [x for x in dset.header["chan_name"] if any(l in x for l in ["ECG", "EKG"])]
    if len(ecg) > 1:
        ecg = choose_best_ecg(ecg, dset, num=1, logger=logger)

    return choose_best_eeg(eeg, ecg, dset, num=num, logger=logger)
    

def infer_eog(dset, logger=create_logger('Infer EOG')):
    """
    Use candidate EOG channels and quality-check to return best pair.

    Parameters
    ----------
    dset : object
        The dataset object containing signal data and metadata.
    logger : logging.Logger, optional
        Logger for messages.

    Returns
    -------
    list
        List of selected EOG channel names (ideally left + right).
    """
    eyes = get_candidate_eog(dset, logger=logger)

    if len(eyes) == 0 or len(eyes) > 6:
        logger.error('Unable to determine EOG channels from recording. EOG names must be explicitly specified.')
        logger.info('Check documentation for how to specify channels:')
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
        logger.info('-' * 10)
        return None

    if len(eyes) > 2:
        loc = [x for x in eyes if any(l in x for l in ['l', 'L'])]
        loc = choose_best_eog(loc, dset, 1, logger) if loc else []

        roc = [x for x in eyes if any(l in x for l in ['r', 'R'])]
        roc = choose_best_eog(roc, dset, 1, logger) if roc else []

        eyes = roc + loc

    return eyes


def infer_emg(dset, logger=create_logger('Infer EMG')):
    """
    Use candidate EMG channels and apply quality checks to select the best one.

    Parameters
    ----------
    dset : object
        The dataset object containing signal data and metadata.
    logger : logging.Logger, optional
        Logger for messages.

    Returns
    -------
    list
        List containing the best EMG channel (or None).
    """
    emg = get_candidate_emg(dset, logger=logger)

    if len(emg) == 0 or len(emg) > 6:
        logger.error('Unable to determine EMG channels from recording. EMG names must be explicitly specified.')
        logger.info('Check documentation for how to specify channels:')
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
        logger.info('-' * 10)
        return None

    if len(emg) > 1:
        emg = choose_best_emg(emg, dset, num=1, logger=logger)

    return emg

        
# def infer_eeg(self, dset, filt = None, logger = create_logger('Infer EEG')):
    
#     ''' This function reads a recording file (e.g. edf) and searches for likely
#         EEG channel names. IF the EEG channels are not easly identifiable from
#         the edf, then nothing is returned. IF there are multiple (>2) options,
#         then the channels will be QC'd and the channels with the best quality 
#         will be returned. #NOTE: Central channels will be given highest priority.
#     '''
    
#     h = dset.header
    
#     if filt:
#         if not isinstance(filt, list):
#             raise ValueError('`filt` must be a list')
    
#         eeg = [x for x in h['chan_name'] if any(l in x for l in filt)] 
    
#     else:
#         # 1. Look for central channels first
#         eeg = [x for x in h['chan_name'] if any(l in x for l in ['C1', 'C2', 'C3',
#                                                                  'C4', 'C5', 'C6',
#                                                                  'Cz', 'CZ'])]  
        
#         # 2. If no central channels, look for frontal:
#         if len(eeg) == 0:
#             eeg = [x for x in h['chan_name'] if any(l in x for l in ['F1', 'F2', 'F3',
#                                                                      'F4', 'F5', 'F6',
#                                                                      'Fz', 'FZ'])]          
#         # 3. Look for any channels names beginning with E
#         if len(eeg) == 0:
#             eeg = [x for x in h['chan_name'] if 'E' in x if not 
#                                                     any(l in x for l in ['EMG', 
#                                                                          'ECG',
#                                                                          'EOG',
#                                                                          'Chin',
#                                                                          ])]
               
#     # Check for ECG channel
#     ecg = [x for x in h['chan_name'] if any(l in x for l in ['ECG', 'EKG'])]  
#     if len(ecg) > 1:
#         ecg = self.choose_best_ecg(ecg, dset, num = 1, logger = logger)

    
#     # If still nothing or when all electrode names contain 'E' (e.g. EGI nets) 
#     if len(eeg) == 0 or len(eeg) > 10:
#         logger.error('Unable to determine EEG channels from recording. EEG names must be explicitly specified.')
#         logger.info('Check documentation for how to specify channels:')
#         logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
#         logger.info('-' * 10)
#         return None
#     else:
#         eeg = self.choose_best_eeg(eeg, ecg, dset, 1, logger)
            
#     return eeg 
    
# def infer_eog(self, dset, logger = create_logger('Infer EOG')):
    
#     ''' This function reads a recording file (e.g. edf) and searches for likely
#         EOG channel names. IF the EOG channels are not easly identifiable from
#         the edf, then nothing is returned. IF there are multiple (>2) options,
#         then the channels will be QC'd and the channels with the best quality 
#         will be returned.
#     '''
    
#     h = dset.header
    
#     # First and most obvious EOG name
#     eyes = [x for x in h['chan_name'] if 'EOG' in x]  
    
#     # If EOG are not labelled as 'EOG'
#     if len(eyes) == 0:
#         eyes = [x for x in h['chan_name'] if 'E' in x if not 'EMG' in x 
#                                                       if not 'ECG' in x]    
    
#     # If still nothing or when all electrode names contain 'E' (e.g. EGI nets) 
#     if len(eyes) == 0 or len(eyes) > 6:
#         logger.error('Unable to determine EOG channels from recording. EOG names must be explicitly specified.')
#         logger.info('Check documentation for how to specify channels:')
#         logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
#         logger.info('-' * 10)
#         return None

#     elif len(eyes) > 2:
        
#         loc = [x for x in eyes if any(l in x for l in ['l', 'L'])]
#         loc = self.choose_best_eog(loc, dset, 1, logger)
        
#         roc = [x for x in eyes if any(l in x for l in ['r', 'R'])]
#         roc = self.choose_best_eog(roc, dset, 1, logger)
        
#         eyes = roc + loc
            
#     return eyes    

# def infer_emg(self, dset, logger = create_logger('Infer EMG')):
    
#     ''' This function reads a recording file (e.g. edf) and searches for likely
#         EOG channel names. IF the EOG channels are not easly identifiable from
#         the edf, then nothing is returned. IF there are multiple (>2) options,
#         then the channels will be QC'd and the channels with the best quality 
#         will be returned.
#     '''
    
#     h = dset.header
    
#     # First and most obvious EMG name
#     emg = [x for x in h['chan_name'] if any(l in x for l in ['EMG', 
#                                                              'Chin', 
#                                                              'CHIN'])]  
     
    
#     # If still nothing or when all electrode names contain 'E' (e.g. EGI nets) 
#     if len(emg) == 0 or len(emg) > 6:
#         logger.error('Unable to determine EMG channels from recording. EOG names must be explicitly specified.')
#         logger.info('Check documentation for how to specify channels:')
#         logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
#         logger.info('-' * 10)
#         return None

#     elif len(emg) > 1:
#         emg = self.choose_best_emg(emg, dset, 1, logger)
            
#     return emg  


def choose_best_eeg(chans, ecg, dset, num = 2, logger = create_logger('QC EEG')):
    """
    Choose the best EEG channel(s) based on quality metrics computed using `squid`.

    Parameters
    ----------
    chans : list of str
        List of EEG channel names to evaluate.
    
    dset : object
        The dataset object containing signal data and associated metadata.
    
    num : int, optional, default=1
        The number of top channels to return based on their quality scores.
    
    logger : logging.Logger, optional
        Logger instance for logging any warnings or debug information.
    
    Returns
    -------
    eeg_names : list of str
        List of top `num` EEG channel names based on their quality score.
    """
    ch_dat = dset.read_data(chan = chans)
  
    all_quals = []

    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        
        quals = SQUID.squid_eeg(ch_dat.data[0][i], s_freq, name)
        quals['channel'] = name  # optionally tag with channel name
        
        if quals.loc['time_not_flatline'] > 50:
            logger.warning(f'Channel {name} is mostly flatlined... Review.')
    
        if quals.loc['time_below_200'] < 0.75:
            logger.warning(f"{quals.loc[i,'time_below_200']*100}% of the channel "
                           f"{name} is > 200 µV. This is implausible. Check if "
                           "the signal contains a lot of artefact or has proper "
                           "scaling.")
        
        all_quals.append(quals)
    
    # Combine all into a single DataFrame
    all_quals_df = concat(all_quals, ignore_index = True)  
    
    # Mask only numeric and relevant columns
    numeric_cols = all_quals_df.select_dtypes(include = 'number').columns
    
    # Apply ranking only to those columns
    winners = all_quals_df[numeric_cols].rank(ascending=False)
    
    # Put back the channel names for clarity
    winners['channel'] = all_quals_df['channel']
    
    #winners = DataFrame([all_quals_df[x].rank() for x in all_quals_df.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    eeg_names = [chans[x] for x in idx]
    
    logger.debug(f'Best EEG channels based on auto-QC are: {eeg_names}')
    
    return eeg_names


def choose_best_eog(chans, dset, num = 2, logger = create_logger('QC EOG')):
    """
    Choose the best EOG channel(s) based on quality metrics computed using `squid`.

    Parameters
    ----------
    chans : list of str
        List of EOG channel names to evaluate.
    
    dset : object
        The dataset object containing signal data and associated metadata.
    
    num : int, optional, default=1
        The number of top channels to return based on their quality scores.
    
    logger : logging.Logger, optional
        Logger instance for logging any warnings or debug information.
    
    Returns
    -------
    eog_names : list of str
        List of top `num` EOG channel names based on their quality score.
    """
    ch_dat = dset.read_data(chan = chans)
  
    all_quals = []

    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        
        quals = SQUID.squid_eeg(ch_dat.data[0][i], s_freq, name)
        quals['channel'] = name  # optionally tag with channel name
        all_quals.append(quals)
    
    # Combine all into a single DataFrame
    all_quals_df = concat(all_quals, ignore_index = True)  
    
    # Mask only numeric and relevant columns
    numeric_cols = all_quals_df.select_dtypes(include = 'number').columns
    
    # Apply ranking only to those columns
    winners = all_quals_df[numeric_cols].rank(ascending=False)
    
    # Put back the channel names for clarity
    winners['channel'] = all_quals_df['channel']
    
    #winners = DataFrame([all_quals_df[x].rank() for x in all_quals_df.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    eog_names = [chans[x] for x in idx]
    
    logger.debug(f'Best EOG channels based on auto-QC are: {eog_names}')
    
    return eog_names


def choose_best_emg(chans, dset, num=1, logger=create_logger('QC EMG')):
    """
    Chooses the best EMG channels based on quality metrics computed using `squid_emg`.
    
    Parameters
    ----------
    chans : list of str
        List of channel names to evaluate.
    
    dset : object
        The dataset object containing signal data and associated metadata.
        Assumes `dset.read_data(chan=chans)` can retrieve the EMG signal for each channel.
    
    num : int, optional, default=1
        The number of top channels to return based on their quality scores.
    
    logger : logging.Logger, optional
        Logger instance for logging any warnings or debug information.
    
    Returns
    -------
    best_channels : list of str
        List of `num` top EMG channels, ranked by quality score.
    """
    
    quals = DataFrame(columns=['time_not_flatline', 
                               'stdev', 
                               'power_low', 
                               'power_high', 
                               'broadband_ratio',
                               'snr_proxy', 
                               'cv', 
                               'var_ratio', 
                               'quality_score', 
                               'reasons'
                               ])
    
    emg_dat = dset.read_data(chan=chans)
    
    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][i]
        signal = emg_dat.data[0][i]
        
        # Run the squid_emg function to get quality metrics
        quals_row = SQUID.squid_emg(signal, s_freq, name, logger)
        
        # Append the results to the dataframe
        quals = concat([quals, quals_row], ignore_index=True)
    
    # Rank the channels based on quality score (higher is better)
    winners = quals[['quality_score']].rank(axis=0, ascending=False)
    idx = winners.sum(axis=1).nlargest(num).index.to_list()
    emg_name = [chans[x] for x in idx]
    
    logger.debug(f'Best EMG channels based on auto-QC are: {emg_name}')
    
    return emg_name
        

def choose_best_ecg(chans, dset, num=1, logger=None):
    """
    Choose the best ECG channel(s) based on quality metrics computed using `squid_ecg`.

    Parameters
    ----------
    chans : list of str
        List of ECG channel names to evaluate.
    
    dset : object
        The dataset object containing signal data and associated metadata.
    
    num : int, optional, default=1
        The number of top channels to return based on their quality scores.
    
    logger : logging.Logger, optional
        Logger instance for logging any warnings or debug information.
    
    Returns
    -------
    ecg_names : list of str
        List of top `num` ECG channel names based on their quality score.
    """
    if logger is None:
        logger = create_logger('QC_ECG')

    ecg_dat = dset.read_data(chan=chans)
    results = []

    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][i]
        e = ecg_dat.data[0][i]

        # Compute quality metrics for this channel
        quals = SQUID.squid_ecg(e, s_freq, name, logger)
        results.append(quals)

    # Combine Results into DataFrame
    df_quals = concat(results, ignore_index=True)

    # Rank numeric columns only
    numeric_cols = df_quals.select_dtypes(include='number').columns
    df_ranks = df_quals[numeric_cols].rank(ascending=False)

    # Add overall quality score and rank
    df_quals['quality_score'] = df_ranks.sum(axis=1)
    df_quals['quality_rank'] = df_quals['quality_score'].rank(ascending=False)

    # Select best ECG channel(s)
    top_channels = df_quals.sort_values('quality_score', ascending=False).head(num)
    ecg_names = top_channels['channel'].tolist()

    logger.debug(f'Best ECG channel(s) based on auto-QC: {ecg_names}')

    return ecg_names

    
    
def squid_eeg(data, s_freq, name = None, weights = None, 
          logger = create_logger('SQUID')):
    
    """
        S.Q.U.I.D. – EEG Edition:

        Computes EEG-specific QC metrics and returns a DataFrame row with flags.

        Parameters
        ----------
        data : array-like
            Input timeseries data. Expected to be a nested structure accessible 
            via `ch_dat.data[0][i]`, where `i` indexes the channel. Assumes that 
            associated metadata such as sampling frequency is available globally 
            or via a `dset` object.
            
        Returns
        -------
        quals : pandas.DataFrame
            A dataframe containing the following quality metrics per channel:
                - time_not_flatline : float
                    Proportion of time the signal is not flat (based on moving average).
                - time_above_10 : float
                    Proportion of time the signal amplitude is above 10 µV.
                - time_below_200 : float
                    Proportion of time the signal stays below 200 µV.
                - gini_coeff : float
                    Gini coefficient of signal amplitude distribution.
                - inverse_power_ratio : float
                    Inverse of the ratio of high-frequency to low-frequency power.
                - ecg_artefact : float
                    Measure of ECG contamination based on average peak correlation.
                - ecg_artefact_perc : float
                    Proportion of ECG artefact time with high (r>0.5) correlation.
    
    """

    quals = DataFrame(dtype=float, index=[0])
    reasons = []

    # --- Flatline ---
    e = moving_average_time(abs(data), s_freq, 5)
    flat_ratio = sum(diff(e) == 0) / len(diff(e))
    flat_score = 1 - flat_ratio
    quals.loc[0, 'time_not_flatline'] = flat_score
    quals.loc[0, 'is_flatlined'] = flat_ratio > 0.5
    if flat_ratio > 0.5:
        reasons.append("Flatline > 50%")

    # --- Amplitude constraints ---
    above_10 = logical_and(e >= 0, e <= 10)
    amp_score = round(1 - sum(e > 200) / len(e), 2)
    quals.loc[0, 'time_above_10'] = 1 - sum(above_10) / len(e)
    quals.loc[0, 'time_below_200'] = amp_score
    quals.loc[0, 'has_high_amplitude'] = amp_score < 0.75
    if amp_score < 0.75:
        reasons.append("Signal > 200µV > 25%")

    # --- Gini coefficient ---
    g = gini(e)
    quals.loc[0, 'gini_coeff'] = g

    # --- Power ratio ---
    ratio = transform_signal(data, s_freq, 'moving_power_ratio', 
                             method_opt={'dur': 5,
                                         'freq_narrow': (30, 50),
                                         'freq_broad': (0.5, 5),
                                         'fft_dur': 30,
                                         'step': 15
                                         })
    inv_ratio = 1 / (abs(nanmean(ratio)) + 1e-6)
    quals.loc[0, 'inverse_power_ratio'] = inv_ratio

    # --- ECG contamination ---
    (ecg_artifacts, 
     ecg_template, 
     ecg_corr) = detect_ecg_artifacts(data, sfreq = s_freq)
    (avg_peak, 
     prop_high_corr, 
     total_ecg_area) = compute_ecg_impact_metrics(ecg_artifacts, ecg_corr)
    ecg_score = round(prop_high_corr, 2)
    quals.loc[0, 'ecg_artefact'] = abs(avg_peak)
    quals.loc[0, 'ecg_artefact_perc'] = ecg_score
    quals.loc[0, 'ecg_contaminated'] = prop_high_corr > 0.25
    if prop_high_corr > 0.25:
        reasons.append("ECG contamination > 25%")

    # --- Flags & reasons ---
    quals.loc[0, 'is_bad_channel'] = quals.loc[0, ['is_flatlined', 
                                                   'has_high_amplitude', 
                                                   'ecg_contaminated']].any()
    quals.loc[0, 'reason'] = '; '.join(reasons) if reasons else "OK"
    if name:
        quals.loc[0, 'channel'] = name

    # --- Quality scoring ---
    default_weights = {'time_not_flatline': 0.25,
                       'time_below_200': 0.2,
                       'ecg_clean': 0.2,
                       'low_inv_ratio': 0.2,
                       'gini_balance': 0.15
                       }
    w = weights or default_weights
    total_weight = sum(w.values())

    # Nonlinear transforms
    inv_ratio_scaled = 1 / (1 + exp(-5 * (inv_ratio - 1)))  # sigmoid
    gini_scaled = exp(-10 * (g - 0.5) ** 2)          # bell curve

    quality_score = (w['time_not_flatline'] * flat_score +
                     w['time_below_200'] * amp_score +
                     w['ecg_clean'] * (1 - ecg_score) +
                     w['low_inv_ratio'] * inv_ratio_scaled +
                     w['gini_balance'] * gini_scaled
                     ) / total_weight

    quals.loc[0, 'quality_score'] = round(quality_score, 3)

    # --- Quality label ---
    if quality_score < 0.5:
        label = 'bad'
    elif quality_score < 0.7:
        label = 'medium'
    else:
        label = 'good'
    quals.loc[0, 'quality_label'] = label

    return quals


def squid_emg(signal, s_freq, name=None, weights = None, 
              logger = create_logger('SQUID')):
    """
    S.Q.U.I.D. – EMG Edition: 

    Computes EMG-specific QC metrics and returns a DataFrame row with flags.

    Parameters
    ----------
    signal : array-like
        The raw EMG signal for a given channel, expected to be a 1D array of voltage 
        measurements (in µV).
    
    s_freq : float
        The sampling frequency (Hz) of the signal. Typically derived from the 
        dataset's header or metadata.
    
    name : str, optional
        The name of the channel. Used for logging and reporting purposes.
    
    logger : logging.Logger, optional
        Logger instance for recording warnings or debug information.
    
    Returns
    -------
    quals : pandas.DataFrame
        A dataframe containing the following quality metrics per channel:
            - time_not_flatline : float
                Proportion of time the signal is not flat, based on the difference
                of the moving average (smoothed absolute signal).
            - stdev : float
                Standard deviation of the raw signal, reflecting the signal variability.
            - power_low : float
                Total low-frequency power (20-50 Hz), used as a proxy for signal activity.
            - power_high : float
                Total high-frequency power (70-150 Hz), used to assess the presence of
                muscle bursts or noise.
            - broadband_ratio : float
                Ratio of high-frequency to low-frequency power. High values indicate
                bursts or sharp activity in the signal.
            - snr_proxy : float
                Signal-to-noise ratio, approximated as the ratio of high-frequency power
                to total power.
            - cv : float
                Coefficient of variation of the signal, a measure of burstiness.
            - var_ratio : float
                Variance ratio between maximum windowed variance and mean variance,
                indicating burst characteristics in the signal.
            - quality_score : float
                An overall quality score, calculated as the mean of several key metrics.
            - reasons : str
                Comma-separated string of reasons for signal quality issues (e.g., "flatline",
                "low_stdev", "low_cv").
    
    """

    quals = DataFrame(index=[0], dtype=float)
    reasons = []

    # --- Flatline detection using smoothed rectified signal ---
    e = moving_average_time(signal, s_freq, 5)
    flatness = 1 - sum(diff(e) == 0) / len(e)
    quals.loc[0, 'time_not_flatline'] = flatness

    if flatness < 0.95:
        reasons.append("flatline")
        
    # --- Standard deviation ---
    stdev = nanstd(signal)
    quals.loc[0, 'stdev'] = stdev

    if stdev < 1:
        reasons.append("low_stdev")

    # --- Power spectral density ---
    f, Pxx = periodogram(signal, s_freq)
    low_band = logical_and(f >= 20, f <= 50)
    high_band = logical_and(f >= 70, f <= 150)

    power_low = trapz(Pxx[low_band], f[low_band])
    power_high = trapz(Pxx[high_band], f[high_band])
    total_power = trapz(Pxx, f)

    quals.loc[0, 'power_low'] = power_low
    quals.loc[0, 'power_high'] = power_high
    quals.loc[0, 'broadband_ratio'] = power_high / (power_low + 1e-6)
    quals.loc[0, 'snr_proxy'] = power_high / (total_power + 1e-6)

    if power_high < 1e-2:
        reasons.append("low_high_freq_power")

    # --- Coefficient of variation (burstiness) ---
    cv = variation(signal)
    quals.loc[0, 'cv'] = cv

    if cv < 0.1:
        reasons.append("low_cv")

    # --- Variance ratio: burst detection proxy ---
    win_size = int(s_freq * 2)  # 2-second windows
    n_windows = len(signal) // win_size
    windowed_vars = [var(signal[i*win_size:(i+1)*win_size]) for i in range(n_windows)]

    if windowed_vars:
        var_ratio = max(windowed_vars) / (mean(windowed_vars) + 1e-6)
        quals.loc[0, 'var_ratio'] = var_ratio
        if var_ratio < 2:
            reasons.append("low_burstiness")
    else:
        quals.loc[0, 'var_ratio'] = nan
        reasons.append("short_signal")
        if logger: logger.warning(f"{name}: Too short for variance-based burstiness")
 
    # --- Quality scoring ---
    default_weights = {'time_not_flatline': 0.25,
                       'stdev_scaled': 0.2,
                       'broadband_scaled': 0.15,
                       'cv_scaled': 0.15,
                       'var_ratio_scaled': 0.15,
                       'snr_scaled': 0.1
                       }
    w = weights or default_weights
    
    # --- Nonlinear transforms / scaling ---
    # These help map all features to ~[0, 1], where higher = better
    quals['stdev_scaled'] = 1 / (1 + exp(-1 * (quals['stdev'] - 2)))  # sigmoid centered ~2µV
    quals['broadband_scaled'] = exp(-0.1 * quals['broadband_ratio'])  # penalize very high ratios
    quals['cv_scaled'] = 1 / (1 + exp(-10 * (quals['cv'] - 0.2)))     # sigmoid centered around healthy CV
    quals['var_ratio_scaled'] = 1 / (1 + exp(-2 * (quals['var_ratio'] - 2)))  # ideal burstiness > 2
    quals['snr_scaled'] = 1 / (1 + exp(-10 * (quals['snr_proxy'] - 0.05)))   # sigmoid tuned to low SNR
    
    # Ensure no NaNs interfere
    score_fields = list(default_weights.keys())
    valid_scores = [w[k] * quals.loc[0, k] for k in score_fields if not isnan(quals.loc[0, k])]
    valid_weights = [w[k] for k in score_fields if not isnan(quals.loc[0, k])]
    
    # --- Final summary score (mean of key features) ---
    quals.loc[0, 'quality_score'] = (sum(valid_scores) / sum(valid_weights) 
                                     if valid_scores else nan)

    if name:
        quals.loc[0, 'channel'] = name
        
    # --- Quality label ---
    if quals.loc[0, 'quality_score'] < 0.5:
        label = 'bad'
    elif quals.loc[0, 'quality_score'] < 0.7:
        label = 'medium'
    else:
        label = 'good'
    quals.loc[0, 'quality_label'] = label

    # --- Add flags as string list ---
    quals.loc[0, 'reasons'] = ", ".join(reasons) if reasons else ""

    return quals
    
def squid_ecg(data, s_freq, name=None, logger=None):
    """
    S.Q.U.I.D. – ECG Edition: 
    
    Computes a series of quality metrics for ECG signals, including:
    - Heart rate (BPM)
    - Standard deviation
    - Flatline detection
    - Power spectral density (PSD) analysis

    Parameters
    ----------
    data : array-like
        ECG signal data for a given channel.
    
    s_freq : float
        Sampling frequency (Hz) of the ECG signal.
    
    name : str, optional
        The name of the channel (for logging purposes).
    
    logger : logging.Logger, optional
        Logger instance for logging any warnings or debug information.

    Returns
    -------
    quals : pandas.DataFrame
        A dataframe containing the following quality metrics per channel:
            - hr_qual : int
                Heart rate quality (0, 1, 2).
            - stdev : float
                Standard deviation of the absolute ECG signal.
            - time_not_flatline : float
                Proportion of time the signal is not flat.
            - PSD_qual : int
                Quality of the power spectral density (0, 1, 2).
    """
    quals = DataFrame(data=None, columns=['hr_qual', 'stdev', 'time_not_flatline', 'PSD_qual'])

    # Heart Rate Check
    beats = detect_heartbeats(data, s_freq)
    hr = len(beats) / (len(data) / (s_freq * 60))

    if 40 < hr < 120:
        quals['hr_qual'] = 2
        if logger:
            logger.debug(f'Heart rate ({round(hr)} bpm) is plausible.')
    elif 120 <= hr < 200:
        quals['hr_qual'] = 1
        if logger:
            logger.warning(f'Heart rate ({round(hr)} bpm) is plausible but high.')
    else:
        quals['hr_qual'] = 0
        if logger:
            logger.warning(f'Heart rate ({round(hr)} bpm) is NOT plausible.')

    # Standard Deviation
    quals['stdev'] = nanstd(abs(data))

    # Flatline Check
    quals['time_not_flatline'] = 1 - (sum(diff(data) == 0) / len(diff(data)))

    # PSD Quality
    f_p, Pxx_spec = periodogram(data, s_freq)
    Pxx_spec_sm = concatenate((
        zeros(2000),
        moving_average(Pxx_spec, 4000),
        zeros(1999)
    ))
    top = int((len(data) / s_freq) * 20)  # up to ~20 Hz
    peaks = find_peaks(Pxx_spec_sm[:top], prominence=800)[0]

    if len(peaks) > 4:
        quals['PSD_qual'] = 2
    elif len(peaks) > 2:
        quals['PSD_qual'] = 1
    else:
        quals['PSD_qual'] = 0
        if logger:
            logger.warning(f'PSD looks unusual for channel: {name}')
    
    return quals


def gather_qc_reports(qc_root, modality='eeg'):
    """
    Gather all QC reports for a given modality into a single DataFrame.
    
    Parameters
    ----------
    qc_root : str or Path
        Root path to the derivatives/QC folder.
    modality : str
        Modality to search for (e.g., 'eeg', 'emg').
        
    Returns
    -------
    pd.DataFrame
        Combined DataFrame of all qc_report_<modality>.csv files.
    """
    qc_root = Path(qc_root)
    all_reports = []
    
    # Look for files like sub-*/ses-*/qc_report_<modality>.csv
    for csv_file in qc_root.glob("sub-*/ses-*/sub-*_ses-*qc_report_{}.csv".format(modality)):
        try:
            df = read_csv(csv_file, sep=None, engine='python')  # Handles tabs or commas
            df['file_path'] = str(csv_file)
            all_reports.append(df)
        except Exception as e:
            print(f"Could not read {csv_file}: {e}")
    
    if all_reports:
        combined_df = concat(all_reports, ignore_index=True)
        return combined_df
    else:
        return DataFrame()  # Return empty dataframe if none found
    
       
        
#%% Quality Control Checks 

def gini(x, w = None):
    
    ''' Calculates the Gini coefficient:
                    a measure of statistical dispersion calculated by comparing 
                    the actual Lorenz curve to the diagonal line of equality.
    '''

    x = asarray(x)
    if w is not None:
        w = asarray(w)
        sorted_indices = argsort(x)
        sorted_x = x[sorted_indices]
        sorted_w = w[sorted_indices]
        # Force float dtype to avoid overflows
        cumw = cumsum(sorted_w, dtype=float)
        cumxw = cumsum(sorted_x * sorted_w, dtype=float)
        return (sum(cumxw[1:] * cumw[:-1] - cumxw[:-1] * cumw[1:]) / 
                (cumxw[-1] * cumw[-1]))
    else:
        sorted_x = sort(x)
        n = len(x)
        cumx = cumsum(sorted_x, dtype=float)
        # The above formula, with all weights equal to 1 simplifies to:
        return (n + 1 - 2 * sum(cumx) / cumx[-1]) / n


def compute_ecg_impact_metrics(ecg_artifacts, ecg_corr, percentile_threshold=99):
    """
    Compute 3 summary metrics for ECG artifact contamination.

    Parameters:
    - ecg_artifacts: array of sample indices with detected ECG events
    - ecg_corr: full correlation time series from template matching
    - percentile_threshold: for "high correlation" cutoff (default=99th percentile)

    Returns:
    - avg_peak_corr: mean ECG correlation at detected artifact peaks (Metric 1)
    - high_corr_ratio: % of samples with ECG-like correlation above threshold (Metric 2)
    - ecg_area: total correlation "area" across detected ECG events (Metric 3)
    """

    # Metric 1: Avg correlation at detected artifact peaks
    avg_peak_corr = mean(ecg_corr[ecg_artifacts])

    # Metric 2: Ratio of signal strongly affected by ECG
    threshold = 0.5
    high_corr_ratio = mean(ecg_corr > threshold)

    # Metric 3: Area under the correlation curve at ECG peaks
    ecg_area = sum(ecg_corr[ecg_artifacts])

    return avg_peak_corr, high_corr_ratio, ecg_area    



## Other functions for SQUID:   
def moving_average_time(a, s_freq, win=5):
    n = int(win*s_freq*60)
    ret = cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def moving_average(a, win=100):
    n = int(win)
    ret = cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n    
    



"""Packages containing sleep events detection

"""
from .fish import FISH, extract_data_error, extract_pac_data, extract_event_data
from .surf import erp_analysis, insert_virtual_markers, marker_to_annot
from .seasnakes import seasnakes, swordfish
from .whales import whales